/**
  *  \file client/widgets/configstorage.cc
  *  \brief WConfigStorageControl widget
  */

#include <algorithm>
#include "client/widgets/configstorage.h"
#include "cpluslib/assert.h"
#include "ui/listbox.h"
#include "ui/uiglobal.h"
#include "ui/util.h"
#include "util/nls.h"
#include "util/strtab.h"

static const char* const sources[] = {
    N_("(Multiple files)"),
    N_("Default value"),
    N_("System configuration file"),
    N_("User configuration file"),
    N_("Game configuration file"),
};
CTASSERT(ConfigOption::srcDefault == 0);
CTASSERT(ConfigOption::srcSystem == 1);
CTASSERT(ConfigOption::srcUser == 2);
CTASSERT(ConfigOption::srcGame == 3);


/** Constructor.
    \param autosave true: change gets effective immediately; false: needs save() */
WConfigStorageControl::WConfigStorageControl(bool autosave)
    : btn_select(0, "S", 's'),
      options(),
      autosave(autosave),
      status(-1)
{
    add(btn_select, GfxRect(0, 0, 20, 20));
    btn_select.addCall(this, &WConfigStorageControl::onSelect);
    btn_select.setState(st_Disabled, true);
}

WConfigStorageControl::~WConfigStorageControl()
{ }

void
WConfigStorageControl::getLayoutInfo(LayoutInfo& info)
{
    info.min_size = info.pref_size =
        GfxPoint(getTextWidth(*fonts[FONT_NORMAL], StaticTranslatedStringTable(sources, countof(sources))) + 25,
                 std::max(20, font_heights[FONT_NORMAL]));
    info.flags = li_FlexibleH;
}

void
WConfigStorageControl::drawContent(GfxCanvas& can)
{
    GfxContext ctx;
    ctx.useCanvas(can);
    ctx.useFont(*fonts[FONT_NORMAL]);
    ctx.useColorScheme(getSkin());
    ctx.setColor(UIColor::tc_Static);
    ctx.setSolidBackground();
    ctx.setTextAlign(0, 1);

    GfxRect r = getExtent();
    r.consumeX(25);
    if (status >= 0 && status < int(countof(sources))) {
        outTextF(ctx, r, _(sources[status]));
    } else {
        drawBackground(ctx, r);
    }
    can.update(getExtent());
}

void
WConfigStorageControl::onResize(GfxRect oldSize, bool byLayout)
{
    UIWidget::onResize(oldSize, byLayout);
    btn_select.setExtent(GfxRect(getExtent().x, getExtent().y, 20, getExtent().h), true);
}

/** Remove all options. Use addOption() to add another set of options to control. */
void
WConfigStorageControl::clear()
{
    options.clear();
    status = -1;
    btn_select.setState(st_Disabled, true);
    drawWidget();
}

/** Add an option.
    \param opt Option to observe */
void
WConfigStorageControl::addOption(ConfigOption& opt)
{
    // Remember
    options.push_back(&opt);
    btn_select.setState(st_Disabled, false);

    // Update status
    int thisOptionStatus = opt.getSource() + 1;
    if (status < 0 || status == thisOptionStatus) {
        status = thisOptionStatus;
    } else {
        status = 0;
    }
    drawWidget();
}

/** Handle button press. This displays the possible locations and allows the user to choose one. */
void
WConfigStorageControl::onSelect()
{
    StringList menu;
    menu.add(ConfigOption::srcUser + 1, _(sources[ConfigOption::srcUser+1]));
    menu.add(ConfigOption::srcGame + 1, _(sources[ConfigOption::srcGame+1]));

    UIStandardListbox lb(0, menu.getCount(), getTextWidth(*fonts[FONT_NORMAL], menu)+20, menu);
    if (lb.doStandardMenu(btn_select)) {
        status = lb.getKey();
        if (autosave) {
            save();
        }
        sig_change.raise();
        drawWidget();
    }
}

/** Make user changes effective. */
void
WConfigStorageControl::save()
{
    // Status -1 means no option is selected.
    // Status 0 means multiple files, which cannot be the result of a change.
    // So we only need to handle the positive values.
    if (status > 0) {
        ConfigOption::Source src = ConfigOption::Source(status - 1);
        for (std::size_t i = 0; i < options.size(); ++i) {
            options[i]->setSource(src);
        }
    }
}
