/**
  *  \file game/planet-form.cc
  *  \brief Planet-related Formulas
  *
  *  (c) 2001-2012 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <cmath>
#include "game/planet.h"
#include "util/nls.h"
#include "cpluslib/misc.h"
#include "game/config.h"
#include "game/formula.h"
#include "game/host.h"
#include "game/planet-int.h"

static int
computeHappinessTarget(int happy)
{
    if (happy >= 100) {
        return 0;
    } else {
        int target = (100 - happy) / 6;
        if (happy < 100 && target == 0)
            target = 1;
        return target;
    }
}

static inline double
pdiv(int a, int b)
{
    if (host.isPHost())
        return double(a) / b;
    else
        return a / b;
}

/** Maximum number of structures.
    \param clans      clans on planet
    \param threshold  below threshold, growth is linear; above it's rootic :-) */
int
GPlanet::getMaxStructuresFormula(int32 clans, int threshold)
{
    if (clans <= threshold)
        return clans;
    else
        /* the rounding function is actually ERnd, but that never
           makes a difference; sqrt(INT) never ends in .5 exactly */
        return threshold + int(std::sqrt(double(clans - threshold)) + 0.5);
}

/** Maximum number of structures.
    \param ps     Structure to query
    \param clans  Number of clans to assume
    \return Maximum allowed structures of that type */
int
GPlanet::getMaxStructures(TPlanetaryStructure ps, int32 clans) const
{
    switch (ps) {
     case str_Factories:
        return getMaxStructuresFormula(clans, 100);
     case str_Mines:
        return getMaxStructuresFormula(clans, 200);
     case str_Defense:
        return getMaxStructuresFormula(clans, 50);
     case str_BaseDefense:
        return base_info ? config.MaximumDefenseOnBase(base_info->data.owner) : 0;
    }
    return 0;
}

/** Colonists change.
    \param tax   desired tax rate
    \param mifa  mines plus factories on planet */
mn32_t
GPlanet::getColonistChange(int tax, int mifa) const
{
    /* can we compute that? */
    if (!getColonists().isKnown() || !isOwnerKnown() || !getTemperature().isKnown())
        return mn32_t();

    double common = 1000 - 80*tax - std::sqrt(double(getColonists()));
    bool   crystal = (raceId(getOwner()) == 7) && config.CrystalsPreferDeserts();
    if (host.isPHost()) {
        if (crystal)
            return mn32_t(int32(common - mifa/3.0 - (100-getTemperature())/0.66) / 100);
        else
            return mn32_t(int32(common - mifa/3.0 - std::abs(getTemperature()-50)/0.33) / 100);
    } else {
        if (crystal)
            return mn32_t(int32(common - (mifa/3 + 3*(100-getTemperature()))) / 100);
        else
            return mn32_t(int32(common - (mifa/3 + 3*std::abs(getTemperature()-50))) / 100);
    }
}

/** Colonists happiness change on this planet.
    \returns change on this planet in current situation */
mn32_t
GPlanet::getColonistChange() const
{
    if (!getColonistTax().isKnown() || !getStructures(str_Mines).isKnown() || !getStructures(str_Factories).isKnown())
        return mn32_t();
    else
        return getColonistChange(getColonistTax(), getStructures(str_Mines) + getStructures(str_Factories));
}

/** Get colonist tax amount, as requested. This returns the amount we want
    from the colonists, not the amount we'll finally get.
    \pre isOwnerKnown()
    \pre getColonists().isKnown() */
int32
GPlanet::getColonistDue(const int tax) const
{
    // Note that these formulas differ in rounding only. PHost uses
    // `Round', THost uses `ERnd' aka `I-don't-care-how-it-rounds'.
    ASSERT(isOwnerKnown() && getColonists().isKnown());
    const int rate = config.ColonistTaxRate(getOwner());
    if (host.isPHost()) {
        return radiv(radiv(getColonists() * tax, 1000) * rate, 100);
    } else {
        return rdivaddl(rdivaddl(getColonists() * tax, 1000, 0) * rate, 100, 0);
    }
}

/** Get colonist tax income.
    \param tax     [in] tax rate
    \param rem_inc [out] this is set to the amount of taxes which we can
    still get from the natives (regarding MaxPlanetaryIncome)
    \returns megacredits collected.
    \pre isOwnerKnown()
    \pre getColonists().isKnown() */
int32
GPlanet::getColonistDueLimited(const int tax, int32& rem_inc) const
{
    const int32 due = getColonistDue(tax), max = config.MaxPlanetaryIncome(getOwner());
    if (due < max) {
        rem_inc = max - due;
        return due;
    } else {
        rem_inc = 0;
        return max;
    }
}

/** Maximum population on planet, for a race.
    \param race player number
    \returns maximum population in clans. */
mp32_t
GPlanet::getMaxSupportedColonists(int player) const
{
    bool crystal = (raceId(player) == 7 && config.CrystalsPreferDeserts());
    int32 limit;
    if (host.isPHost()) {
        if (!config.ClimateLimitsPopulation())
            return mp32_t(250000);
        if (!getTemperature().isKnown())
            return mp32_t();
        if (crystal) {
            if (config.CrystalSinTempBehavior())
                if (getTemperature() >= 15)
                    return mp32_t(int32(100000 * std::sin(getTemperature() * PI / 100)));
                else
                    return mp32_t(3 + getTemperature() * config.MaxColTempSlope() / 100);
            else
                return mp32_t(1000 * getTemperature());
        }

        if (getTemperature() >= 85)
            limit = config.MaxColTempSlope() * (100-getTemperature()) / 100 + 1;
        else if (getTemperature() <= 14)
            limit = config.MaxColTempSlope() * getTemperature() / 100 + 3;
        else
            limit = int32(100000 * std::sin(getTemperature() * PI / 100));
    } else {
        if (!config.ClimateLimitsPopulation())
            return mp32_t(100000);
        if (!getTemperature().isKnown())
            return mp32_t();
        if (crystal)
            return mp32_t(1000 * getTemperature());

        if (getTemperature() >= 85)
            limit = 2 * (100 - getTemperature()) + 1;
        else if (getTemperature() <= 14)
            limit = 2 * getTemperature() + 3;
        else
            limit = int32(100000 * std::sin((100 - getTemperature()) * 0.0314) + 0.5);
    }

    if (raceId(player) == 10 && getTemperature() <= 19 && limit < 90000)
        limit = 90000;
    if ((raceId(player) == 4 || raceId(player) >= 9) && getTemperature() >= 84 && limit < 60)
        limit = 60;

    return mp32_t(limit);
}

/* Natives */

/** Native happiness change.
    \param tax   desired tax rate
    \param mifa  mines plus factories on planet */
mn32_t
GPlanet::getNativeChange(int tax, int mifa) const
{
    /* formula is correct for PHost and THost */
    if (!getNativeGovernment().isKnown() || !getNatives().isKnown())
        return mn32_t();

    int change =
        int32(500 + 50 * getNativeGovernment()
              - 85 * tax
              - pdiv(mifa, 2)
              - std::sqrt(double(getNatives())))
        / 100;
    if (getNativeRace().isKnown() && getNativeRace() == nr_Avian)
        change += 10;
    return mn32_t(change);
}

/** Native happiness change.
    \returns happiness change for current situation. */
mn32_t
GPlanet::getNativeChange() const
{
    if (!getNativeTax().isKnown() || !getStructures(str_Mines).isKnown() || !getStructures(str_Factories).isKnown())
        return mn32_t();
    else
        return getNativeChange(getNativeTax(), getStructures(str_Mines) + getStructures(str_Factories));
}

/** Get native tax amount, as requested, for this planet. This returns
    the amount we're asking from them, not what we'll get.
    \pre native race, government, population, and planet owner known */
int32
GPlanet::getNativeDue(int tax) const
{
    ASSERT(getNativeRace().isKnown() && getNativeGovernment().isKnown());
    ASSERT(getNatives().isKnown() && isOwnerKnown());

    return getNativeDue(tax, getNativeRace(), getNativeGovernment(),
                        getNatives(), getOwner());
}

/** Get native tax amount, as requested, parameterized. This returns
    the amount we're asking from them, not what we'll get.
    \param tax    tax rate
    \param race   native race
    \param gov    government factor (SPI)
    \param pop    native clans
    \param owner  owner (for production rates) */
int32
GPlanet::getNativeDue(int tax, int race, int gov, int32 pop, int owner)
{
    int32 due;
    if (host.isPHost()) {
        due = radiv(radiv(tax * gov * pop, 5000) * config.NativeTaxRate(owner), 100);
    } else {
        due = rdivaddl(rdivaddl(tax * gov * pop, 5000, 0) * config.ColonistTaxRate(owner), 100, 0);
    }
    if (race == nr_Insectoid)
        return 2*due;
    else
        return due;
}

/** Get native tax amount, limited. Limits the tax income by income
    limit and available colonists.
    \param tax     tax rate
    \param rem_inc remaining allowed income; output of getColonistDueLimited(). */
int32
GPlanet::getNativeDueLimited(int tax, int32 rem_inc) const
{
    /* amorphs don't pay */
    if (getNativeRace() == nr_Amorphous)
        return 0;

    /* cyborgs can only tax to 20 */
    int limit = host.getNativeTaxRateLimit(getOwner());
    if (tax > limit)
        tax = limit;

    /* normal formulas here */
    int32 due = getNativeDue(tax);
    int32 colos = getColonists();

    if (host.isPHost()) {
        if (getNativeRace() == nr_Insectoid)
            colos *= 2;
        colos = radiv(colos * config.NativeTaxRate(getOwner()), 100);
    } else {
        colos = colos * config.ColonistTaxRate(getOwner()) / 100;
        if (getNativeRace() == nr_Insectoid)
            colos *= 2;
    }

    if (due > colos)
        due = colos;
    if (due > rem_inc)
        due = rem_inc;
    return due;
}

/** Get Bovinoid supply contribution. Contribution is not yet limited by
    colonists. */
int32
GPlanet::getBovinoidSupplyContribution() const
{
    // FIXME: we now return 0 if the value is unknown
    if (getNativeRace().isKnown() && isOwnerKnown() && getNatives().isKnown())
        if (getNativeRace() == nr_Bovinoid)
            return getBovinoidSupplyContribution(getNatives(), getOwner());
        else
            return 0;
    else
        return 0;
}

/** Get Bovinoid supply contribution. This returns the amount of
    supplies finally collected. */
int32
GPlanet::getBovinoidSupplyContributionLimited() const
{
    // FIXME: the same problem as in getBovinoidSupplyContribution(int,int)
    // appears here. We *could* add a workaround here (compute totals,
    // compute factory contribution, compute difference), but is it
    // really worth to do that for an accuracy of +/-1 with a config
    // option no-one uses?
    //      -- update: it's going to be used in PList 2.5, so someday
    //         it should be supported
    if (!getNativeRace().isKnown() || !getColonists().isKnown())
        return 0;

    if (getNativeRace() == nr_Bovinoid) {
        int32 limit = host.isPHost() ? getColonists() * config.ProductionRate(getOwner()) / 100 : getColonists();
        int32 due = getBovinoidSupplyContribution();
        if (due < limit)
            return due;
        else
            return limit;
    } else {
        return 0;
    }
}

/** Get Amorphous colonist breakfast. Returns the number of clans
    eaten by our lovely amorphous natives when they have a happiness
    as specified. */
int32
GPlanet::getAmorphousBreakfast(int happy)
{
    if (host.isPHost()) {
        if (happy >= 70)
            return 5;
        else if (happy >= 50)
            return 20;
        else
            return 40;
    } else {
        if (happy > 90)
            return 5;
        else
            return 95 - happy;
    }
}

/** Get Bovinoid supply contribution, parameterized. */
int32
GPlanet::getBovinoidSupplyContribution(int32 pop, int owner)
{
    // FIXME: for ultimate accuracy, we have to consider factories as
    // well. PHost does "(bovi + factories) * ProductionRate", not
    // "bovi * ProductionRate + factories * ProductionRate".
    if (host.isPHost())
        return (pop / 100) * config.ProductionRate(owner) / 100;
    else
        return pop / 100;
}

/** Get mining capacity.
    \param index Mineral index
    \param mines Number of mines on planet
    \pre density is known */
int
GPlanet::getMiningCapacity(int index, int mines) const
{
    // Gather inputs
    int density = getOreDensity(index);

    int mining_rate;
    if (isOwnerKnown()) {
        mining_rate = config.RaceMiningRate(getOwner());
    } else {
        mining_rate = 100;
    }

    int reptile_factor;
    if (getNativeRace().isKnown()
        && getNativeRace() == nr_Reptilian
        && (!getNatives().isKnown() || getNatives() > 0))
    {
        reptile_factor = 2;
    } else {
        reptile_factor = 1;
    }

    // Host-dependant formula
    if (!host.isPHost()) {
        return rdivaddl(rdivaddl(density * int32(mines), 100, 0) * mining_rate, 100, 0) * reptile_factor;
    } else {
        int add = (host.isPHostRoundingMiningResults() ? 50 : 0);
        return (((density * int32(mining_rate) + add) / 100) * reptile_factor * mines + add) / 100;
    }
}

/** Get native "safe tax" rate.
    \param mifa Assume this many mines and factories
    \return Safe tax value, or unknown if preconditions not met */
mp16_t
GPlanet::getNativeSafeTax(int mifa) const
{
    // Validate inputs
    if (!isOwnerKnown()
        || !getColonists().isKnown()
        || !getNativeRace().isKnown()
        || !getNatives().isKnown()
        || !getNativeGovernment().isKnown()
        || !getNativeHappiness().isKnown()
        || getNatives() <= 0)
    {
        return mp16_t();
    }

    // Compute result
    int taxlimit = host.getNativeTaxRateLimit(getOwner());
    int tax = taxlimit;
    if (getNativeHappiness() < 70 || getNativeRace() == nr_Amorphous) {
        // Use tax 0 for unhappy natives or Amorphs
        tax = 0;
    } else {
        // Figure out maximum tax rate yielding a usable happiness:
        int target = computeHappinessTarget(getNativeHappiness());
        while (tax > 0 && getNativeChange(tax, mifa) < target)
            --tax;
    }

    // If higher tax rate produces the same happiness change, use that.
    // This applies when the happiness change goal cannot be reached,
    // so let's use a tax rate that gets some income instead of 0%.
    while (tax < taxlimit && getNativeChange(tax, mifa) == getNativeChange(tax+1, mifa))
        ++tax;

    // If lower tax rates produce the same income, use that.
    // Note that the original code tried (and failed) to handle the relation
    // between colonist tax, native tax, and MaxPlanetaryIncome here. Doing
    // this correctly means handling the (assumed) colonist tax rate here,
    // which would complicate matters too much for my taste. This code matters
    // in two places: where safe-tax income hits MaxPlanetaryIncome (rare),
    // and where population is low enough to make rounding effects matter.
    int32_t limit = 0x7FFFFFFF;
    int32_t income = getNativeDueLimited(tax, limit);
    while (tax > 0 && getNativeDueLimited(tax-1, limit) == income)
        --tax;

    return mp16_t(tax);
}

/** Get colonist "safe tax" rate.
    \param mifa Assume this many mines and factories
    \return Safe tax value, or unknown if preconditions not met */
mp16_t
GPlanet::getColonistSafeTax(int mifa) const
{
    // Validate inputs
    if (!isOwnerKnown()
        || !getColonists().isKnown()
        || !getColonistTax().isKnown()
        || !getColonistHappiness().isKnown()
        || !getTemperature().isKnown())
    {
        return mp16_t();
    }

    // Compute result
    int taxlimit = host.getColonistTaxRateLimit(getOwner());
    int tax = taxlimit;
    if (getColonistHappiness() < 70) {
        // Use tax 0 for unhappy colonists
        tax = 0;
    } else {
        // Figure out maximum tax rate yielding a usable happiness:
        int target = computeHappinessTarget(getColonistHappiness());
        while (tax > 0 && getColonistChange(tax, mifa) < target)
            --tax;
    }

    // If higher tax rate produces the same happiness change, use that.
    // This applies when the happiness change goal cannot be reached,
    // so let's use a tax rate that gets some income instead of 0%.
    while (tax < taxlimit && getColonistChange(tax, mifa) == getColonistChange(tax+1, mifa))
        ++tax;

    // If lower tax rates produce the same income, use that:
    int32 limit;
    int32 income = getColonistDueLimited(tax, limit);
    while (tax > 0 && getColonistDueLimited(tax-1, limit) == income)
        --tax;

    return mp16_t(tax);
}

/** Get native "base tax" rate. The base tax assumes a fixed mines/factories
    count to make results comparable between planets.
    \param happyTarget Happiness target (0=base tax, -30=max tax)
    \return tax rate, unknown if preconditions not met */
mp16_t
GPlanet::getNativeBaseTax(int happyTarget) const
{
    // Check preconditions
    if (!isOwnerKnown()
        || !getNativeRace().isKnown()
        || !getNatives().isKnown()
        || !getNativeGovernment().isKnown())
    {
        return mp16_t();
    }

    // Acquire data
    int nid = getNativeRace();
    int32_t natives = getNatives();
    int spi = getNativeGovernment();
    int player = getOwner();

    // No taxes if natives wouldn't pay anyway, or there are none
    if (natives == 0 || nid == nr_Amorphous) {
        return mp16_t(0);
    }

    // Compute result
    const int mifa = 100;
    int tax = host.getNativeTaxRateLimit(player);
    while (tax > 0) {
        /* Host 3.22.040 formula:
           Trunc(1000.0 - sqrt(Planet.Natives) - 85 * Planet.NativeTax
                        - (Planet.Mines + .Factories) DIV 2
                        - 50*(10-Planet.NativeGov))
              DIV 100
           change:=5 + Trunc(spi/2 - (Sqrt(Natives) + 85*LONGINT(tax) + 100/2) / 100);
           change:=5 + Trunc(50*spi - (85*LONGINT(tax) + 50) - Sqrt(Natives)) DIV 100;
           since MIFA is even, the same formula can be used for PHost and THost. */
        int change = int32(500 - (mifa/2) + (50*spi) - (85*tax) - std::sqrt(double(natives))) / 100;
        if (nid == nr_Avian)
            change += 10;
        if (change >= happyTarget)
            break;
        --tax;
    }
    return mp16_t(tax);
}

/** Compute cost for a tech level. The cost is normalized to an arbitrary, unspecified
    basis. Since tech costs are always used to go from X to Y, you'd subtract two
    getBaseTechCost()s.
    \param tech Tech level
    \return Cost */
int32_t
GPlanet::getBaseTechCost(int tech) const
{
    return getBaseTechCost(getOwner(), tech);
}

/** Compute cost for a tech level. The cost is normalized to an arbitrary, unspecified
    basis. Since tech costs are always used to go from X to Y, you'd subtract two
    getBaseTechCost()s.
    \param player Player to compute this for
    \param tech Tech level
    \return Cost */
int32_t
GPlanet::getBaseTechCost(int player, int tech)
{
    /* Going from tech i to i+1 costs 100*i mc, thus going from 0 to i
       costs 100*\sum[0,i] i = 100*(i*(i-1)/2) = 50*i*(i-1).

       Using BaseTechCost, the formula becomes
           BaseTechCost*(i*(i-1)/2)
       Becuase either i or i-1 will be even, the division by two will
       always work out. */
    return config.BaseTechCost(player) * tech * (tech-1) / 2;
}
