/**
  *  \file int/data.cc
  *  \brief Interpreter: Data Structures
  *
  *  This defines the data model used in the interpreter. Most
  *  persistent data is stored in IntDataSegment's.
  *
  *  (c) 2001-2012 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <algorithm>
#include "int/data.h"
#include "cpluslib/assert.h"
#include "io/io.h"
#include "io/memstream.h"
#include "io/streams.h"

/** Create blank data segment. */
IntDataSegment::IntDataSegment()
{ }

/** Destroy. */
IntDataSegment::~IntDataSegment()
{ }

/** Make index accessible. Makes sure that \c values can be indexed
    with \c index. */
void
IntDataSegment::makeIndexAccessible(uint32 index)
{
    if (index >= values.size()) {
        values.reserve(index+1);
        while (index >= values.size()) {
            values.push_back(0);
        }
    }
}

/** Fetch element. Returns index \c index, or 0 if that index has never
    been written. */
IntValue*
IntDataSegment::operator[](uint32 index) const
{
    if (index < values.size())
        return values[index];
    else
        return 0;
}

/** Set element by copying.
    \param index Index to store
    \param value Value to store, may be 0. Will be copied (IntValue::clone()) */
void
IntDataSegment::set(uint32 index, IntValue* value)
{
    setNew(index, IntValue::cloneOf(value));
}

/** Set element to newly-allocated value.
    \param index Index to store
    \param value Value to store, may be 0. The value will become owned by
                 the IntDataSegment object. If the assignment results in
                 an exception, it will be deleted. */
void
IntDataSegment::setNew(uint32 index, IntValue* value)
{
    try {
        makeIndexAccessible(index);
    }
    catch (...) {
        delete value;
        throw;
    }
    values.replace_elem(index, value);
}

/** Append value by copying.
    \param value Value to append, may be 0. Will be copied (IntValue::clone()) */
void
IntDataSegment::push(IntValue* value)
{
    values.push_back(IntValue::cloneOf(value));
}

/** Remove last element. */
void
IntDataSegment::pop()
{
    ASSERT(values.size());
    values.pop_back();
}

/** Remove last n elements. */
void
IntDataSegment::pop(uint32_t count)
{
    for (uint32_t i = 0; i < count; ++i)
        pop();
}

/** Get top-of-stack element.
    \param index Element index. 0 (default) returns the last pushed element,
                 1 the one before, and so on.
    \pre index < size() */
IntValue*
IntDataSegment::top(uint32 index) const
{
    ASSERT(index < values.size());
    return values[values.size() - index - 1];
}

/** Swap data segment elements.
    \param index Element index in this data segment.
    \param other Other data segment
    \param otherIndex Element index in other data segment. */
void
IntDataSegment::swap(uint32 index, IntDataSegment& other, uint32 otherIndex)
{
    makeIndexAccessible(index);
    other.makeIndexAccessible(otherIndex);
    values.swap_elems(index, other.values, otherIndex);
}

/** Clear this data segment. Discards all elements. */
void
IntDataSegment::clear()
{
    ptr_vector<IntValue>().swap(values);
}

/** Get number of used slots. Returns the smallest possible number such that all
    slots at higher numbers yield null with operator[]. */
uint32
IntDataSegment::getNumberOfUsedSlots()
{
    uint32 result = values.size();
    while (result > 0 && values[result-1] == 0)
        --result;
    return result;
}

/** Save to stream.
    \param out Stream to store in
    \param slots Number of slots to store */
void
IntDataSegment::save(Stream& out, uint32 slots) const
{
    // Collect headers in one stream, aux data in another
    MemoryStream headers, aux;

    for (uint32 i = 0; i < slots; ++i) {
        // Generate single entry
        IntTagNode sv;
        if (IntValue* v = operator[](i)) {
            v->store(sv, aux);
        } else {
            sv.tag   = sv.Tag_Empty;
            sv.value = 0;
        }

        // Serialize tag node
        storeStructureT(headers, sv);
    }

    // Generate output
    out.writeT(headers.getBuffer().ptr(), headers.getBuffer().getSize());
    out.writeT(aux.getBuffer().ptr(),     aux.getBuffer().getSize());
}
