/**
  *  \file int/vmio.cc
  *  \brief Interpreter: Virtual Machine I/O
  *
  *  A VM file contains a pool of objects, namely
  *  - bytecode objects (BCOs, IntBytecodeObject)
  *  - processes (IntExecutionFrame)
  *  - data (IntHashData, IntArrayData)
  *
  *  Data is stored properly referenced and is not flattened. This is
  *  required to keep suspended call-by-reference intact, for example,
  *  code like this:
  *  <code>
  *     Sub foo(a)
  *       Stop
  *       a("x") := "y"
  *     EndSub
  *     Sub bar()
  *       Local a := NewHash()
  *       foo(a)
  *       Print a("x")
  *     EndSub
  *  </code>
  *  This implies we also have to deal with forward references.
  *
  *  Upon saving, we assign each object an Id. We also build a plan
  *  with a save order that minimizes forward references, by saving
  *  each object's preconditions before the object itself.
  *
  *  Upon loading, when we encounter an object reference without
  *  having seen the object yet, we create a blank object. When the
  *  object finally appears in the file, it is filled. It is an error
  *  for a VM file to contain references to objects it does not
  *  contain, but this error is not (yet) detected; it's harmless
  *  because empty objects are still valid.
  *
  *  (c) 2001-2012 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "int/vmio.h"
#include "cpluslib/assert.h"
#include "int/contextio.h"
#include "int/exec.h"
#include "int/interror.h"
#include "int/process.h"
#include "io/io.h"
#include "io/nullstream.h"
#include "util/except.h"
#include "util/nls.h"
#include "int/structure.h"

/** VM File Format: Object Types. */
enum {
    otyp_Bytecode        = 1,    ///< Bytecode object (procedure).
    otyp_Process         = 2,    ///< Process object.
    otyp_Frame           = 3,    ///< Stack frame object.
    otyp_DataArray       = 4,    ///< Data object: array.
    otyp_DataHash        = 5,    ///< Data object: hash.
    otyp_DataStructValue = 6,    ///< Data object: structure value.
    otyp_DataStructType  = 7     ///< Data object: structure type.
};

/** VM File Format: BCO flags. */
enum {
    bcofl_Procedure = 1,         ///< Set if this is a procedure, clear if it is a function.
    bcofl_Varargs   = 2          ///< Set if this routine takes a variable number of arguments.
};

/** VM File Format: Frame flags. */
enum {
    frfl_WantResult = 1         ///< Set if caller wants a result.
};

/**************************** IntVMSaveContext ***************************/

namespace {
    class SaveObject {
        Stream& s;
        uint32_t obj_header[4]; // type, size, id, nproperties
        uint32_t header_pos;
        uint32_t property_index;
        uint32_t this_property_start;
        std::vector<uint32_t> properties;

        void writeHeader();
     public:
        SaveObject(Stream& s);
        void start(uint32_t type, uint32_t id, uint32_t nprop);
        void end();
        void startProperty(uint32_t count);
        void endProperty();
    };

    inline void
    storeItem(char* p, uint32_t v)
    {
        storeUint32(p, v);
    }

    inline void
    storeItem(char* p, const IntOpcode& op)
    {
        *p++ = op.arg;
        *p++ = op.arg >> 8;
        *p++ = op.minor;
        *p++ = op.major;
    }

    inline void
    storeItem(char* p, IntBCORef bco)
    {
        storeUint32(p, IntVMSaveContext::getCurrentInstance()->addBCO(*bco));
    }

    template<typename T>
    void
    writeArray32(Stream& out, const T& obj)
    {
        enum { CHUNK = 128 };
        char buffer[CHUNK*4];
        uint32_t total = obj.size();
        uint32_t index = 0;
        while (index < total) {
            uint32_t now = total - index;
            if (now > CHUNK)
                now = CHUNK;
            for (uint32_t i = 0; i < now; ++i) {
                storeItem(&buffer[4*i], obj[index + i]);
            }
            out.writeT(buffer, 4*now);
            index += now;
        }
    }
}

SaveObject::SaveObject(Stream& s)
    : s(s)
{ }

void
SaveObject::writeHeader()
{
    std::vector<char> buffer;

    // Header
    buffer.resize(4*4);
    storeUint32(&buffer.front(), obj_header, 4);
    s.writeT(&buffer.front(), buffer.size());

    // Property table
    buffer.resize(4*properties.size());
    storeUint32(&buffer.front(), &properties.front(), properties.size());
    s.writeT(&buffer.front(), buffer.size());
}

void
SaveObject::start(uint32_t type, uint32_t id, uint32_t nprop)
{
    ++nprop;
    obj_header[0] = type;
    obj_header[1] = id;
    obj_header[2] = 0;
    obj_header[3] = nprop;
    header_pos = s.getPos();
    properties.clear();
    properties.resize(2*nprop);
    properties[0] = 0;
    properties[1] = 0;
    property_index = 1;
    writeHeader();
}

void
SaveObject::end()
{
    ASSERT(property_index == obj_header[3]);
    uint32_t end_pos = s.getPos();
    obj_header[2] = end_pos - header_pos - 4*4;
    s.seek(header_pos);
    writeHeader();
    s.seek(end_pos);
}

void
SaveObject::startProperty(uint32_t count)
{
    this_property_start = s.getPos();
    properties[2*property_index] = count;
}

void
SaveObject::endProperty()
{
    properties[2*property_index + 1] = s.getPos() - this_property_start;
    ++property_index;
}

/** Constructor. Makes a blank save context. */
IntVMSaveContext::IntVMSaveContext()
    : prev_instance(save_instance),
      obj_id_counter(0),
      current_process(0)
{
    save_instance = this;
}

/** Destructor. */
IntVMSaveContext::~IntVMSaveContext()
{
    ASSERT(save_instance == this);
    save_instance = prev_instance;
}

/** Save a bytecode object.
    \param out  Stream to write to
    \param p    Bytecode object to save
    \param id   Id to assign this BCO */
void
IntVMSaveContext::saveBCO(Stream& out, const IntBytecodeObject& bco, uint32_t id)
{
    SaveObject so(out);
    so.start(otyp_Bytecode, id, 8);

    // Property 1: header (num_labels, flags, min_args, max_args)
    so.startProperty(0);
    char header[8];
    storeUint16(header+0, bco.is_procedure*bcofl_Procedure + bco.is_varargs*bcofl_Varargs);
    storeUint16(header+2, bco.min_args);
    storeUint16(header+4, bco.max_args);
    storeUint16(header+6, bco.num_labels);
    out.writeT(header, sizeof(header));
    so.endProperty();

    // Property 2: "data" (literals for pushlit, data segment)
    so.startProperty(bco.data.size());
    bco.data.save(out, bco.data.size());
    so.endProperty();

    // Property 3: "names" (names for e.g. pushvar, name list)
    so.startProperty(bco.names.getNumNames());
    bco.names.save(out, bco.names.getNumNames());
    so.endProperty();

    // Property 4: "code" (count = number of instructions, size = 4x count). 32 bit per instruction.
    so.startProperty(bco.code.size());
    writeArray32(out, bco.code);
    so.endProperty();

    // Property 5: "local_names" (predeclared locals, name list)
    so.startProperty(bco.local_names.getNumNames());
    bco.local_names.save(out, bco.local_names.getNumNames());
    so.endProperty();

    // Property 6: "name" (name hint for loading, string)
    so.startProperty(0);
    if (bco.name.size())
        out.writeT(bco.name.data(), bco.name.size());
    so.endProperty();

    // Property 7: "file name" (debug file name, string)
    so.startProperty(0);
    if (bco.file_name.size())
        out.writeT(bco.file_name.data(), bco.file_name.size());
    so.endProperty();

    // Property 8: "line numbers" (count = number of lines, size = 8x count)
    so.startProperty(bco.line_numbers.size()/2);
    writeArray32(out, bco.line_numbers);
    so.endProperty();

    so.end();
}

/** Save a hash object.
    \param out Stream to save to
    \param hash Hash to save
    \param id Id to use */
void
IntVMSaveContext::saveHash(Stream& out, const IntHashData& hash, uint32_t id)
{
    SaveObject so(out);
    so.start(otyp_DataHash, id, 2);

    // Property 1: names
    so.startProperty(hash.names.getNumNames());
    hash.names.save(out, hash.names.getNumNames());
    so.endProperty();

    // Property 2: values
    so.startProperty(hash.content.size());
    hash.content.save(out, hash.content.size());
    so.endProperty();

    so.end();
}

/** Save an array object.
    \param out Stream to save to
    \param array Array to save
    \param id Id to use */
void
IntVMSaveContext::saveArray(Stream& out, const IntArrayData& array, uint32_t id)
{
    SaveObject so(out);
    so.start(otyp_DataArray, id, 2);

    // Property 1: dimensions
    so.startProperty(array.dimensions.size());
    writeArray32(out, array.dimensions);
    so.endProperty();

    // Property 2: content
    so.startProperty(array.content.size());
    array.content.save(out, array.content.size());
    so.endProperty();

    so.end();
}

/** Save a structure type.
    \param out Stream to save to
    \param type Type to save
    \param id Id to use */
void
IntVMSaveContext::saveStructureType(Stream& out, const IntStructureTypeData& type, uint32_t id)
{
    SaveObject so(out);
    so.start(otyp_DataStructType, id, 1);

    // Property 1: name list
    so.startProperty(type.names.getNumNames());
    type.names.save(out, type.names.getNumNames());
    so.endProperty();

    so.end();
}

/** Save a structure value.
    \param out Stream to save to
    \param value Value to save
    \param id Id to use */
void
IntVMSaveContext::saveStructureValue(Stream& out, const IntStructureValueData& value, uint32_t id)
{
    SaveObject so(out);
    so.start(otyp_DataStructValue, id, 2);

    // Property 1: header
    so.startProperty(0);
    char header[4];
    storeUint32(header+0, addStructureType(*value.type));
    out.writeT(header, sizeof(header));
    so.endProperty();

    // Property 2: content
    so.startProperty(value.data.size());
    value.data.save(out, value.data.size());
    so.endProperty();

    so.end();
}

/** Save a stack frame object.
    \param out Stream to save to
    \param fr  Stack frame to save */
void
IntVMSaveContext::saveFrame(Stream& out, const IntExecutionFrame& fr)
{
    SaveObject so(out);

    // We don't actually need the frame_sp here (it will be ignored and
    // reconstructed upon load), but it doesn't hurt.
    so.start(otyp_Frame, fr.frame_sp, 3);

    // Property 1: header
    so.startProperty(0);
    char header[20];
    storeUint32(header+0, addBCO(*fr.bco));
    storeUint32(header+4, fr.pc);
    storeUint32(header+8, fr.context_sp);
    storeUint32(header+12, fr.exception_sp);
    storeUint32(header+16, fr.want_result * frfl_WantResult);
    out.writeT(header, sizeof(header));
    so.endProperty();

    // Property 2: local values (data segment)
    so.startProperty(fr.local_values.size());
    fr.local_values.save(out, fr.local_values.size());
    so.endProperty();

    // Property 3: local names (name list)
    so.startProperty(fr.local_names.getNumNames());
    fr.local_names.save(out, fr.local_names.getNumNames());
    so.endProperty();

    so.end();
}

/** Save process.
    \param out Stream to save to
    \param exc Process to save */
void
IntVMSaveContext::saveProcess(Stream& out, IntExecutionContext& exc)
{
    // Tell users
    IntExecutionContext* old_process = current_process;
    current_process = &exc;

    try {
        // Start the object
        SaveObject so(out);
        so.start(otyp_Process, 0, 6);

        // Property 1: header
        so.startProperty(0);
        char header[4];
        header[0] = exc.getPriority();
        header[1] = exc.getProcessKind();
        storeUint16(&header[2], exc.context_tos);
        out.writeT(header, sizeof(header));
        so.endProperty();

        // Property 2: name (string)
        so.startProperty(0);
        if (exc.process_name.size())
            out.writeT(exc.process_name.data(), exc.process_name.size());
        so.endProperty();

        // Property 3: frames (object array)
        so.startProperty(exc.frames.size());
        for (ptr_vector<IntExecutionFrame>::const_iterator i = exc.frames.begin(); i != exc.frames.end(); ++i)
            saveFrame(out, **i);
        so.endProperty();

        // Property 4: contexts (data segment)
        so.startProperty(exc.contexts.size());
        saveContexts(out, exc.contexts);
        so.endProperty();

        // Property 5: exceptions (counts = number, size = 16xcount)
        so.startProperty(exc.exceptions.size());
        for (ptr_vector<IntExceptionHandler>::const_iterator i = exc.exceptions.begin(); i != exc.exceptions.end(); ++i) {
            char tmp[16];
            storeUint32(tmp+0,  (*i)->frame_sp);
            storeUint32(tmp+4,  (*i)->context_sp);
            storeUint32(tmp+8,  (*i)->value_sp);
            storeUint32(tmp+12, (*i)->pc);
            out.writeT(tmp, sizeof(tmp));
        }
        so.endProperty();

        // Property 6: value stack (data segment)
        so.startProperty(exc.value_stack.size());
        exc.value_stack.save(out, exc.value_stack.size());
        so.endProperty();

        // Finish
        so.end();
        current_process = old_process;
    }
    catch (...) {
        current_process = old_process;
        throw;
    }
}

/** Add bytecode object to serialization. This assigns the BCO an Id
    number which can be used to refer to that BCO. The Id number is
    returned. Adding an object multiple times only saves it once.
    \param bco the bytecode object
    \return Id number */
uint32_t
IntVMSaveContext::addBCO(IntBytecodeObject& bco)
{
    /* Is this item already known? */
    uint32_t& id = obj_to_id[&bco];
    if (id == 0) {
        /* This object is not yet known. Give it an Id. */
        id = ++obj_id_counter;

        /* Save its preconditions. Note that if the BCO indirectly refers
           to itself, the nested addBCO will see that it already has an
           Id (although it is not yet planned) and just re-use that. */
        NullStream ns;
        saveBCO(ns, bco, id);

        /* Remember the plan */
        plan_objs.push_back(&bco);
        plan_types.push_back(otyp_Bytecode);
    }
    return id;
}

/** Add hash object to serialisation. This assigns the hash an
    Id number which can be used to refer to that hash. The Id
    number is returned. Adding an object multiple times only saves
    it once.
    \param hash the hash data
    \return Id number */
uint32_t
IntVMSaveContext::addHash(IntHashData& hash)
{
    /* Is this item already known? */
    uint32_t& id = obj_to_id[&hash];
    if (id == 0) {
        /* This object is not yet known. Give it an Id. */
        id = ++obj_id_counter;

        /* Save its preconditions */
        NullStream ns;
        saveHash(ns, hash, id);

        /* Remember the plan */
        plan_objs.push_back(&hash);
        plan_types.push_back(otyp_DataHash);
    }
    return id;
}

/** Add array object to serialisation. This assigns the array an
    Id number which can be used to refer to that array. The Id
    number is returned. Adding an object multiple times only saves
    it once.
    \param array the array data
    \return Id number */
uint32_t
IntVMSaveContext::addArray(IntArrayData& array)
{
    /* Is this item already known? */
    uint32_t& id = obj_to_id[&array];
    if (id == 0) {
        /* This object is not yet known. Give it an Id. */
        id = ++obj_id_counter;

        /* Save its preconditions */
        NullStream ns;
        saveArray(ns, array, id);

        /* Remember the plan */
        plan_objs.push_back(&array);
        plan_types.push_back(otyp_DataArray);
    }
    return id;
}

/** Add structure type object to serialisation. This assigns the
    type an Id number which can be used to refer to that type.
    The Id number is returned. Adding an object multiple times
    only saves it once.
    \param array the structure type
    \return Id number */
uint32_t
IntVMSaveContext::addStructureType(IntStructureTypeData& type)
{
    /* Is this item already known? */
    uint32_t& id = obj_to_id[&type];
    if (id == 0) {
        /* This object is not yet known. Give it an Id. */
        id = ++obj_id_counter;

        /* Save its preconditions */
        NullStream ns;
        saveStructureType(ns, type, id);

        /* Remember the plan */
        plan_objs.push_back(&type);
        plan_types.push_back(otyp_DataStructType);
    }
    return id;
}

/** Add structure value object to serialisation. This assigns the
    value an Id number which can be used to refer to it. The Id
    number is returned. Adding an object multiple times only saves
    it once.
    \param array the structure value
    \return Id number */
uint32_t
IntVMSaveContext::addStructureValue(IntStructureValueData& value)
{
    /* Is this item already known? */
    uint32_t& id = obj_to_id[&value];
    if (id == 0) {
        /* This object is not yet known. Give it an Id. */
        id = ++obj_id_counter;

        /* Save its preconditions */
        NullStream ns;
        saveStructureValue(ns, value, id);

        /* Remember the plan */
        plan_objs.push_back(&value);
        plan_types.push_back(otyp_DataStructValue);
    }
    return id;
}

/** Add process object to serialization. Note that, unlike bytecode, a process
    can be saved many times. Loading will then create multiple copies of it.
    \param proc The process object */
void
IntVMSaveContext::addProcess(IntExecutionContext& proc)
{
    /* Prepare it by saving into a temporary memory stream: this
       causes its preconditions (=BCOs) to be saved */
    NullStream ns;
    saveProcess(ns, proc);

    /* Remember the plan */
    plan_objs.push_back(&proc);
    plan_types.push_back(otyp_Process);
}

/** Save all pending objects.
    \param out Stream to save to */
void
IntVMSaveContext::save(Stream& out)
{
    ASSERT(plan_objs.size() == plan_types.size());
    for (uint32_t i = 0; i < plan_objs.size(); ++i) {
        switch (plan_types[i]) {
         case otyp_Bytecode:
            saveBCO(out, *static_cast<IntBytecodeObject*>(plan_objs[i]), obj_to_id[plan_objs[i]]);
            break;
         case otyp_Process:
            saveProcess(out, *static_cast<IntExecutionContext*>(plan_objs[i]));
            break;
         case otyp_DataArray:
            saveArray(out, *static_cast<IntArrayData*>(plan_objs[i]), obj_to_id[plan_objs[i]]);
            break;
         case otyp_DataHash:
            saveHash(out, *static_cast<IntHashData*>(plan_objs[i]), obj_to_id[plan_objs[i]]);
            break;
         case otyp_DataStructType:
            saveStructureType(out, *static_cast<IntStructureTypeData*>(plan_objs[i]), obj_to_id[plan_objs[i]]);
            break;
         case otyp_DataStructValue:
            saveStructureValue(out, *static_cast<IntStructureValueData*>(plan_objs[i]), obj_to_id[plan_objs[i]]);
            break;
         default:
            ASSERT(0);
        }
    }
}

/**************************** IntVMLoadContext ***************************/

class IntVMLoadContext::ObjectLoader {
    Stream& s;
    SubStream sub;
    uint32_t obj_type, obj_id, obj_size, obj_nprop;
    uint32_t prop_next;
    uint32_t prop_count;
    uint32_t current_property;
    uint32_t obj_next;
    std::vector<uint32_t> prop;

    void checkObjSize(uint32_t needed);
 public:
    ObjectLoader(Stream& s)
        : s(s), sub(s), obj_next(s.getPos())
        { }
    bool readObject();
    uint32_t readProperty();

    Stream& getPropertyStream()
        { return sub; }
    uint32_t getPropertyCount() const
        { return prop_count; }
    uint32_t getObjectType() const
        { return obj_type; }
    uint32_t getObjectId() const
        { return obj_id; }
};

void
IntVMLoadContext::ObjectLoader::checkObjSize(uint32_t needed)
{
    if (needed > obj_size) {
        throw FileFormatException(s, _("Invalid size"));
    }
    obj_size -= needed;
}

bool
IntVMLoadContext::ObjectLoader::readObject()
{
    /* Read header */
    std::vector<char> header;
    header.resize(4*4);
    s.seek(obj_next);
    int n = s.read(&header.front(), header.size());
    if (n != int(header.size())) {
        if (n == 0)
            return false;
        throw FileTooShortException(s);
    }

    obj_type  = getUint32(&header[0]);
    obj_id    = getUint32(&header[4]);
    obj_size  = getUint32(&header[8]);
    obj_nprop = getUint32(&header[12]);
    obj_next += 4*4 + obj_size;

    /* Validate */
    checkObjSize(obj_nprop * 8);

    /* Read property headers */
    header.resize(obj_nprop * 8);
    s.readT(&header.front(), header.size());
    prop.clear();
    prop.reserve(2*obj_nprop);
    for (uint32_t i = 0; i < 2*obj_nprop; ++i)
        prop.push_back(getUint32(&header[4*i]));

    /* Initialize properties and skip first one */
    prop_next = s.getPos();
    current_property = 0;
    readProperty();
    return true;
}

uint32_t
IntVMLoadContext::ObjectLoader::readProperty()
{
    /* Do we have another property? */
    if (2*current_property >= prop.size())
        return 0;

    /* Check property */
    uint32_t prop_size;
    uint32_t prop_id;
    prop_count = prop[2*current_property];
    prop_size  = prop[2*current_property + 1];
    prop_id    = current_property++;
    checkObjSize(prop_size);

    /* Initialize content */
    sub.setRange(prop_next, prop_size);
    sub.seek(0);
    prop_next += prop_size;

    return prop_id;
}

namespace {
    inline void
    fetchItem(const char* p, std::vector<uint32_t>& out)
    {
        out.push_back(getUint32(p));
    }

    inline void
    fetchItem(const char* p, std::vector<IntBCORef>& out)
    {
        IntBCORef bco = IntVMLoadContext::getCurrentInstance()->getBCO(getUint32(p));
        if (!bco)
            throw std::runtime_error(_("Invalid object reference"));
        out.push_back(bco);
    }

    inline void
    fetchItem(const char* p, std::vector<IntOpcode>& out)
    {
        IntOpcode o;
        o.arg   = getUint16(p);
        o.minor = p[2];
        o.major = p[3];
        out.push_back(o);
    }

    template<typename T>
    void
    loadArray32(Stream& s, uint32_t count, T& out)
    {
        enum { CHUNK = 128 };
        char buffer[CHUNK*4];
        uint32_t index = 0;
        while (index < count) {
            uint32_t now = count - index;
            if (now > CHUNK)
                now = CHUNK;
            s.readT(buffer, 4*now);
            for (uint32_t i = 0; i < now; ++i) {
                fetchItem(&buffer[4*i], out);
            }
            index += now;
        }
    }

    void loadString(Stream& s, string_t& out)
    {
        char buffer[128];
        int n;
        while ((n = s.read(buffer, sizeof(buffer))) > 0)
            out.append(buffer, n);
    }
}

/** Constructor. Makes a blank instance. */
IntVMLoadContext::IntVMLoadContext()
    : prev_instance(load_instance),
      current_process(0)
{
    load_instance = this;
}

/** Destructor. */
IntVMLoadContext::~IntVMLoadContext()
{
    ASSERT(load_instance == this);
    load_instance = prev_instance;
}

/** Load bytecode object.
    \param ldr Object loader that has just read the object header
    \param id Id of object */
void
IntVMLoadContext::loadBCO(ObjectLoader& ldr, uint32_t id)
{
    /* Note: when implementing the merge-loaded-BCO-with-existing-identical
       optimisation, we must know whether this is the first instance of this
       BCO (optimisation applicable), or whether there already was a forward
       reference. The simplest way would be to duplicate getBCO here. */
    IntBCORef obj = getBCO(id);
    while (uint32_t id = ldr.readProperty()) {
        switch (id) {
         case 1:
            // Header
            {
                char header[8];
                int n = ldr.getPropertyStream().read(header, sizeof(header));
                if (n >= 2) {
                    uint16_t flags = getUint16(header+0);
                    obj->is_procedure = (flags & bcofl_Procedure) != 0;
                    obj->is_varargs   = (flags & bcofl_Varargs) != 0;
                }
                if (n >= 4) {
                    obj->min_args = getUint16(header+2);
                }
                if (n >= 6) {
                    obj->max_args = getUint16(header+4);
                }
                if (n >= 8) {
                    obj->num_labels = getUint16(header+6);
                }
            }
            break;

         case 2:
            // "data" (literals for pushlit, data segment)
            obj->data.load(ldr.getPropertyStream(), ldr.getPropertyCount());
            break;

         case 3:
            // "names" (names for e.g. pushvar, name list)
            obj->names.load(ldr.getPropertyStream(), ldr.getPropertyCount());
            break;

         case 4:
            // "code" (count = number of instructions, size = 4x count). 32 bit per instruction.
            loadArray32(ldr.getPropertyStream(), ldr.getPropertyCount(), obj->code);
            break;

         case 5:
            // "local_names" (predeclared locals, name list)
            obj->local_names.load(ldr.getPropertyStream(), ldr.getPropertyCount());
            break;

         case 6:
            // "name" (name hint for loading, string)
            loadString(ldr.getPropertyStream(), obj->name);
            break;

         case 7:
            // "file name" (debug file name, string)
            loadString(ldr.getPropertyStream(), obj->file_name);
            break;

         case 8:
            // "line numbers" (count = number of lines, size = 8x count)
            loadArray32(ldr.getPropertyStream(), ldr.getPropertyCount()*2, obj->line_numbers);
            break;

         default:
            break;
        }
    }
}

/** Load hash object.
    \param ldr Object loader that has just read the object header
    \param id Id of object */
void
IntVMLoadContext::loadHash(ObjectLoader& ldr, uint32_t id)
{
    Ptr<IntHashData> hash = getHash(id);
    while (uint32_t id = ldr.readProperty()) {
        switch (id) {
         case 1:
            // names
            hash->names.load(ldr.getPropertyStream(), ldr.getPropertyCount());
            break;

         case 2:
            // values
            hash->content.load(ldr.getPropertyStream(), ldr.getPropertyCount());
            break;

         default:
            break;
        }
    }
}

/** Load array object.
    \param ldr Object loader that has just read the object header
    \param id Id of object */
void
IntVMLoadContext::loadArray(ObjectLoader& ldr, uint32_t id)
{
    Ptr<IntArrayData> array = getArray(id);
    while (uint32_t id = ldr.readProperty()) {
        switch (id) {
         case 1:
            // Dimensions. Since these can be used to do evil things, we do not
            // read them directly into the object, but into a temporary buffer
            // where we validate them by using the public API.
            {
                std::vector<uint32_t> dims;
                loadArray32(ldr.getPropertyStream(), ldr.getPropertyCount(), dims);
                for (uint32_t i = 0; i < dims.size(); ++i) {
                    if (!array->addDimension(int32_t(dims[i]))) {
                        throw FileFormatException(ldr.getPropertyStream(), _("Invalid array"));
                    }
                }
            }
            break;

         case 2:
            // values
            array->content.load(ldr.getPropertyStream(), ldr.getPropertyCount());
            break;

         default:
            break;
        }
    }
}

/** Load structure value.
    \param ldr Object loader that has just read the object header
    \param id Id of object */
void
IntVMLoadContext::loadStructureValue(ObjectLoader& ldr, uint32_t id)
{
    Ptr<IntStructureValueData> value = getStructureValue(id);
    while (uint32_t id = ldr.readProperty()) {
        switch (id) {
         case 1:
            // Header
            {
                char header[4];
                int n = ldr.getPropertyStream().read(header, sizeof(header));
                if (n >= 4) {
                    value->type = getStructureType(getUint32(header));
                }
            }
            break;
         case 2:
            // content
            value->data.load(ldr.getPropertyStream(), ldr.getPropertyCount());
            break;

         default:
            break;
        }
    }
}

/** Load structure type.
    \param ldr Object loader that has just read the object header
    \param id Id of object */
void
IntVMLoadContext::loadStructureType(ObjectLoader& ldr, uint32_t id)
{
    Ptr<IntStructureTypeData> type = getStructureType(id);
    while (uint32_t id = ldr.readProperty()) {
        switch (id) {
         case 1:
            // names
            type->names.load(ldr.getPropertyStream(), ldr.getPropertyCount());
            break;

         default:
            break;
        }
    }
}

/** Load process. The process will be created in runnable state.
    \param ldr ObjectLoader that has just read the object header */
void
IntVMLoadContext::loadProcess(ObjectLoader& ldr)
{
    IntExecutionContext& proc = createProcess(string_t());

    // remove contexts created by IntExecutionContext constructor
    while (proc.contexts.size()) {
        proc.contexts.pop_back();
    }

    IntExecutionContext* old_process = current_process;
    current_process = &proc;
    uint16_t loaded_context_tos = 0;
    try {
        while (uint32_t id = ldr.readProperty()) {
            switch (id) {
             case 1:
                // header
                {
                    char hdr[4];
                    int n = ldr.getPropertyStream().read(hdr, sizeof(hdr));
                    if (n >= 1) {
                        proc.process_priority = uint8_t(hdr[0]);
                        handlePriorityChange(proc);
                    }
                    if (n >= 2) {
                        switch (hdr[1]) {
                         case IntExecutionContext::pkShipTask:
                         case IntExecutionContext::pkPlanetTask:
                         case IntExecutionContext::pkBaseTask:
                            proc.process_kind = IntExecutionContext::ProcessKind(hdr[1]);
                            break;
                         default:
                            break;
                        }
                    }
                    if (n >= 4) {
                        loaded_context_tos = getUint16(&hdr[2]);
                    }
                }
                break;

             case 2:
                // name (string)
                loadString(ldr.getPropertyStream(), proc.process_name);
                break;

             case 3:
                // frames (object array)
                loadFrames(proc, ldr.getPropertyStream(), ldr.getPropertyCount());
                break;

             case 4:
                // contexts (data segment)
                loadContexts(ldr.getPropertyStream(), ldr.getPropertyCount(), proc.contexts);
                break;

             case 5:
                // exceptions (counts = number, size = 16xcount)
                for (uint32_t i = 0, end = ldr.getPropertyCount(); i < end; ++i) {
                    char frame[4*4];
                    ldr.getPropertyStream().readT(frame, sizeof(frame));

                    IntExceptionHandler* ex = new IntExceptionHandler();
                    ex->frame_sp   = getUint32(frame+0);
                    ex->context_sp = getUint32(frame+4);
                    ex->value_sp   = getUint32(frame+8);
                    ex->pc         = getUint32(frame+12);
                    proc.exceptions.push_back(ex);
                }
                break;

             case 6:
                // value stack (data segment)
                proc.value_stack.load(ldr.getPropertyStream(), ldr.getPropertyCount());
                break;
            }
        }
    }
    catch (...) {
        /* If loading fails with an exception, make sure that the process will not run */
        current_process = old_process;
        proc.process_state = proc.psTerminated;
        throw;
    }
    if (loaded_context_tos <= proc.contexts.size()) {
        /* Since context-TOS cannot be out of range, we do not install it
           without checking, but we cannot check it before we have loaded
           the whole process. */
        proc.context_tos = loaded_context_tos;
    }
    current_process = old_process;
}

/** Load stack frames.
    \param proc  Process to load stack frames into
    \param s     Stream to read from
    \param count Number of stack frames to read */
void
IntVMLoadContext::loadFrames(IntExecutionContext& proc, Stream& s, uint32_t count)
{
    ObjectLoader ldr(s);
    proc.frames.reserve(count);
    while (count-- > 0) {
        // Read frame object
        if (!ldr.readObject())
            throw FileFormatException(s, _("Invalid frame"));
        if (ldr.getObjectType() != otyp_Frame)
            throw FileFormatException(s, _("Invalid frame type"));

        // Read frame content
        IntExecutionFrame* frame = new IntExecutionFrame();
        proc.frames.push_back(frame);
        frame->frame_sp = proc.frames.size()-1;
        while (uint32_t id = ldr.readProperty()) {
            char tmp[5*4];
            uint32_t flags;
            int n;
            switch (id) {
             case 1:
                // header
                n = ldr.getPropertyStream().read(tmp, sizeof(tm));
                // bco_id is mandatory
                if (n < 4)
                    throw FileFormatException(s, _("Invalid frame"));
                frame->bco          = getBCO(getUint32(tmp+0));
                frame->pc           = getUint32(tmp+4);
                frame->context_sp   = getUint32(tmp+8);
                frame->exception_sp = getUint32(tmp+12);

                flags = getUint32(tmp+16);
                if (flags & frfl_WantResult) {
                    frame->want_result = true;
                }
                break;

             case 2:
                // local values (data segment)
                frame->local_values.load(ldr.getPropertyStream(), ldr.getPropertyCount());
                break;

             case 3:
                // local names (name list)
                frame->local_names.load(ldr.getPropertyStream(), ldr.getPropertyCount());
                break;
            }
        }

        // Validate
        if (!frame->bco) {
            throw FileFormatException(s, _("Invalid frame"));
        }
    }
}

/** Get bytecode object by Id
    \param id Id to check
    \return bytecode object */
IntBCORef
IntVMLoadContext::getBCO(uint32_t id)
{
    IntSubroutineValue* sv = bco_map[id];
    if (sv == 0) {
        sv = new IntSubroutineValue(new IntBytecodeObject());
        bco_map.insert(id, sv);
    }
    return sv->getBytecodeObject();
}

/** Get hash object by Id
    \param id Id to check
    \return hash data */
Ptr<IntHashData>
IntVMLoadContext::getHash(uint32_t id)
{
    IntHash* hv = hash_map[id];
    if (hv == 0) {
        hv = new IntHash(new IntHashData());
        hash_map.insert(id, hv);
    }
    return hv->getData();
}

/** Get array object by Id
    \param id Id to check
    \return array data */
Ptr<IntArrayData>
IntVMLoadContext::getArray(uint32_t id)
{
    IntArray* av = array_map[id];
    if (av == 0) {
        av = new IntArray(new IntArrayData());
        array_map.insert(id, av);
    }
    return av->getData();
}

/** Get structure object by Id
    \param id Id to check
    \return structure data */
Ptr<IntStructureValueData>
IntVMLoadContext::getStructureValue(uint32_t id)
{
    IntStructureValue* sv = struct_value_map[id];
    if (sv == 0) {
        // Create structure with a dummy type. This guarantees that all structures
        // actually have a type, even if the VM file is broken and doesn't create one.
        sv = new IntStructureValue(new IntStructureValueData(new IntStructureTypeData()));
        struct_value_map.insert(id, sv);
    }
    return sv->getValue();
}

/** Get structure type object by Id
    \param id Id to check
    \return structure type data */
Ptr<IntStructureTypeData>
IntVMLoadContext::getStructureType(uint32_t id)
{
    IntStructureType* sv = struct_type_map[id];
    if (sv == 0) {
        sv = new IntStructureType(new IntStructureTypeData());
        struct_type_map.insert(id, sv);
    }
    return sv->getType();
}

/** Make stack frame context by Id. This is used to deserialize stack
    frame contexts. It only works while actually deserializing a process.
    \param id Id read from file
    \return IntExecutionFrameContext object; 0 if those are not allowed here */
IntContext*
IntVMLoadContext::makeFrameContext(uint32_t id) const
{
    if (current_process == 0 || id >= current_process->frames.size())
        return 0;
    return new IntExecutionFrameContext(*current_process->frames[id]);
}

/** Load VM file.
    \param s Stream to read from
    \param acceptProcesses true to accept saved processes (=VM file),
    false to accept only bytecode (=object file) */
void
IntVMLoadContext::load(Stream& s, bool acceptProcesses)
{
    ObjectLoader ldr(s);
    while (ldr.readObject()) {
        if (ldr.getObjectType() == otyp_Process && acceptProcesses) {
            loadProcess(ldr);
        } else if (ldr.getObjectType() == otyp_Bytecode) {
            loadBCO(ldr, ldr.getObjectId());
        } else if (ldr.getObjectType() == otyp_DataArray) {
            loadArray(ldr, ldr.getObjectId());
        } else if (ldr.getObjectType() == otyp_DataHash) {
            loadHash(ldr, ldr.getObjectId());
        } else if (ldr.getObjectType() == otyp_DataStructValue) {
            loadStructureValue(ldr, ldr.getObjectId());
        } else if (ldr.getObjectType() == otyp_DataStructType) {
            loadStructureType(ldr, ldr.getObjectId());
        } else {
            throw FileFormatException(s, _("Unexpected object"));
        }
    }
}
