/**
  *  \file reshack/painter.cc
  *  \brief Resource Editor: Bitmap Editor
  *
  *  (c) 2001-2012 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "reshack/painter.h"
#include "gfx/palette.h"
#include "ui/uiglobal.h"
#include "reshack/colorsel.h"


RHPainter::RHPainter(Ptr<GfxPixmap> pixmap, ColorMode color_mode)
    : pixmap(pixmap),
      backup(),                 // initialized on demand
      tool(0),
      pixmap_context(),
      last_pos(0, 0),
      mouse_is_down(false),
      mouse_color_slot(0),
      scale(0),
      color_mode(color_mode)
{
    colors[0] = COLOR_WHITE;
    colors[1] = COLOR_BLACK;
    colors[2] = FC_HALF;
    pixmap_context.setColor(colors[0]);
}

RHPainter::~RHPainter()
{ }

/** Handle mouse click.
    \param pt point, in world coordinates
    \param btn mouse button status */
void
RHPainter::handleMouse(GfxPoint pt, int btn)
{
    if (!tool) {
        /* we don't have a tool, so ignore this event */
    } else if (btn != 0) {
        if (!mouse_is_down) {
            /* mouse pressed. */
            if (!backup || backup->getSize() != pixmap->getSize())
                backup = pixmap->clone();
            else
                blitIndexed(*backup, *pixmap, 0, 0);

            /* colors */
            mouse_color_slot = (btn & mb_Middle) != 0 ? 2 : (btn & mb_Right) != 0 ? 1 : 0;
            pixmap_context.useCanvas(*pixmap);
            pixmap_context.setColor(colorIndexToImageColor(pixmap, colors[mouse_color_slot]));
            tool->click(pixmap_context, pt, colorIndexToImageColor(pixmap, colors[!mouse_color_slot]));
            drawWidget(false);
        } else {
            /* mouse still pressed. */
            if (tool->needsPreview())
                blitIndexed(*pixmap, *backup, 0, 0);
            tool->drag(pixmap_context, pt);
            updateColor();
            drawWidget(false);
        }
    } else {
        if (mouse_is_down) {
            /* mouse released. */
            if (tool->needsPreview())
                blitIndexed(*pixmap, *backup, 0, 0);
            tool->release(pixmap_context, pt);
            updateColor();
            drawWidget(false);
        }
    }
    mouse_is_down = (btn != 0);
    last_pos      = pt;
}

void
RHPainter::moveOrigin(int dx, int dy)
{
    Ptr<GfxPixmap> c = pixmap->clone();
    c->tileAnchored(*pixmap, GfxRect(*pixmap), dx, dy, 0);
    drawWidget();
}

GfxPoint
RHPainter::worldToScreen(GfxPoint pt)
{
    return GfxPoint((pt.x + extent.x) << scale, (pt.y + extent.y) << scale);
}

GfxPoint
RHPainter::screenToWorld(GfxPoint pt)
{
    return GfxPoint((pt.x - extent.x) >> scale, (pt.y - extent.y) >> scale);
}

/** Update color[mouse_color_slot] from the context, if needed. */
void
RHPainter::updateColor()
{
    uint32_t c = pixmap_context.getRawColor();
    switch (color_mode) {
     case StandardPaletteColor:
     case ForeignPaletteColor:
     case GrayscaleColor:
        // 1:1 map
        colors[mouse_color_slot] = c;
        sig_color_change.raise();
        break;

     case TrueColor:
        // Locate color in palette
        for (int i = 0; i < 256; ++i) {
            if (colorIndexToImageColor(pixmap, i) == c) {
                colors[mouse_color_slot] = i;
                sig_color_change.raise();
                break;
            }
        }
        break;
    }
}

void
RHPainter::drawContent(GfxCanvas& can)
{
    int adj = 0;
    if (scale == 0) {
        pixmap->blit(can, extent.x, extent.y, GfxRect(0, 0, extent.w, extent.h));
    } else {
        /* Figure out size of pixels */
        const GfxPoint size = pixmap->getSize();
        int pixel = 1<<scale;
        int dx = extent.x;
        int dy = extent.y;
        if (scale > 1) {
            /* At 4x and higher, draw grid */
            --pixel;
            for (int y = 0; y <= size.y; ++y) {
                can.drawHLine(GfxPoint(dx, dy + (y<<scale)), (size.x<<scale), standard_colors[COLOR_DARK], SOLID_LINE, GFX_ALPHA_OPAQUE);
            }
            for (int x = 0; x <= size.x; ++x) {
                can.drawVLine(GfxPoint(dx + (x<<scale), dy), (size.y<<scale), standard_colors[COLOR_DARK], SOLID_LINE, GFX_ALPHA_OPAQUE);
            }
            ++dx;
            ++dy;
            ++adj;
        }

        if (color_mode == TrueColor) {
            /* Truecolor: screen has same pixel format */
            for (int y = 0; y < size.y; ++y) {
                for (int x = 0; x < size.x; ++x) {
                    drawSolidBar(can, GfxRect(dx + (x<<scale), dy + (y<<scale), pixel, pixel), pixmap->getPixel(GfxPoint(x, y)));
                }
            }
        } else if (color_mode == ForeignPaletteColor || ui_screen_bits > 8) {
            /* Foreign palette -or- Palettized, and true-color screen. */
            for (int y = 0; y < size.y; ++y) {
                for (int x = 0; x < size.x; ++x) {
                    uint8_t c = pixmap->getPixel(GfxPoint(x, y));
                    drawSolidBar(can, GfxRect(dx + (x<<scale), dy + (y<<scale), pixel, pixel),
                                 SDL_MapRGB(ui_screen->getSystemSurface()->format,
                                            pixmap->getSystemSurface()->format->palette->colors[c].r,
                                            pixmap->getSystemSurface()->format->palette->colors[c].g,
                                            pixmap->getSystemSurface()->format->palette->colors[c].b));
                }
            }
        } else {
            /* Palettized, and screen is palettized. No mapping needed. */
            for (int y = 0; y < size.y; ++y) {
                for (int x = 0; x < size.x; ++x) {
                    drawSolidBar(can, GfxRect(dx + (x<<scale), dy + (y<<scale), pixel, pixel),
                                 pixmap->getPixel(GfxPoint(x, y)));
                }
            }
        }
    }
    int r = (pixmap->getWidth() << scale) + adj;
    int b = (pixmap->getHeight() << scale) + adj;
    if (r < extent.w) {
        drawSolidBar(can, GfxRect(extent.x + r, extent.y, extent.w - r, extent.h), standard_colors[COLOR_BLACK]);
        if (b < extent.h)
            drawSolidBar(can, GfxRect(extent.x, extent.y + b, r, extent.h - b), standard_colors[COLOR_BLACK]);
    } else {
        if (b < extent.h)
            drawSolidBar(can, GfxRect(extent.x, extent.y + b, extent.w, extent.h - b), standard_colors[COLOR_BLACK]);
    }

    /* Aux lines */
    if (r < extent.w) {
        int len = extent.w - r;
        int left = scale ? 0 : r;
        if (len > 20)
            len = 20;
        for (std::vector<int>::iterator i = aux_lines[false].begin(); i != aux_lines[false].end(); ++i)
            can.drawHLine(GfxPoint(extent.x + left, extent.y + (*i << scale)), r + len - left, standard_colors[COLOR_GREENBLACK], SOLID_LINE, GFX_ALPHA_OPAQUE);
    }
    if (b < extent.h) {
        int len = extent.h - b;
        int top = scale ? 0 : b;
        if (len > 20)
            len = 20;
        for (std::vector<int>::iterator i = aux_lines[true].begin(); i != aux_lines[true].end(); ++i)
            can.drawVLine(GfxPoint(extent.x + (*i << scale), extent.y + top), b + len - top, standard_colors[COLOR_GREENBLACK], SOLID_LINE, GFX_ALPHA_OPAQUE);
    }
    
    can.update(extent);
}

bool
RHPainter::handleEvent(const UIEvent& e, bool second_pass)
{
    if (e.type == ev_Mouse) {
        /* If mouse is inside, process it. Also consume all mouse events while mouse is down,
           to track movement outside the widget. */
        if (isInside(GfxPoint(e.mouse.x, e.mouse.y)) || mouse_is_down) {
            setSelectedWidget();
            handleMouse(screenToWorld(GfxPoint(e.mouse.x, e.mouse.y)), e.mouse.b);
            return true;
        }
    } else if (e.type == ev_Key) {
        if (e.key.code == SDLK_ESCAPE && mouse_is_down) {
            setSelectedWidget();
            if (backup) {
                blitIndexed(*pixmap, *backup, 0, 0);
                drawWidget(false);
            }
            mouse_is_down = false;
            return true;
        } else if (e.key.code == ss_Shift + SDLK_RIGHT) {
            setSelectedWidget();
            moveOrigin(1-pixmap->getWidth(), 0);
            return true;
        } else if (e.key.code == ss_Shift + SDLK_LEFT) {
            setSelectedWidget();
            moveOrigin(-1, 0);
            return true;
        } else if (e.key.code == ss_Shift + SDLK_UP) {
            setSelectedWidget();
            moveOrigin(0, -1);
            return true;
        } else if (e.key.code == ss_Shift + SDLK_DOWN) {
            setSelectedWidget();
            moveOrigin(0, 1-pixmap->getHeight());
            return true;
        } else if (e.key.code == 'h') {
            pixmap->flipHorizontal();
            drawWidget();
            return true;
        } else if (e.key.code == 'v') {
            pixmap->flipVertical();
            drawWidget();
            return true;
        }
    }
    return UIWidget::handleEvent(e, second_pass);
}

void
RHPainter::getLayoutInfo(LayoutInfo& info)
{
    info.flags = li_Flexible;
    info.min_size = info.pref_size = pixmap->getSize();
}

void
RHPainter::onStateChange(int flags, bool enable)
{
    if ((flags & st_Selected) && !enable && mouse_is_down)
        handleMouse(last_pos, 0);
    UIWidget::onStateChange(flags, enable);
}

void
RHPainter::addAuxLine(int n, bool onXAxis)
{
    for (std::vector<int>::iterator i = aux_lines[onXAxis].begin(); i != aux_lines[onXAxis].end(); ++i)
        if (*i == n)
            return;
    aux_lines[onXAxis].push_back(n);
    drawWidget();
}

void
RHPainter::removeAuxLine(int n, bool onXAxis)
{
    for (std::vector<int>::size_type i = 0; i != aux_lines[onXAxis].size(); ++i) {
        if (aux_lines[onXAxis][i] == n) {
            aux_lines[onXAxis][i] = aux_lines[onXAxis].back();
            aux_lines[onXAxis].pop_back();
            drawWidget();
        }
    }
}

void
RHPainter::setTool(RHTool* tool)
{
    if (mouse_is_down) {
        setSelectedWidget();
        handleMouse(last_pos, 0);
    }
    this->tool = tool;
}

RHTool*
RHPainter::getTool() const
{
    return tool;
}

void
RHPainter::setZoom(int scale)
{
    if (mouse_is_down) {
        setSelectedWidget();
        handleMouse(last_pos, 0);
    }
    this->scale = scale;
    drawWidget();
}

int
RHPainter::getZoom() const
{
    return scale;
}

void
RHPainter::setColor(bool bg, uint32_t color)
{
    colors[bg] = color;
}

uint32_t
RHPainter::getColor(bool bg) const
{
    return colors[bg];
}

ColorMode
RHPainter::getColorMode() const
{
    return color_mode;
}

Ptr<GfxPixmap>
RHPainter::getPixmap() const
{
    return pixmap;
}

void
RHPainter::setPixmap(Ptr<GfxPixmap> pix)
{
    pixmap = pix;
    drawWidget();
}


GfxPoint
RHPainter::getSize() const
{
    return pixmap->getSize();
}

void
RHPainter::setSize(GfxPoint pt)
{
    if (pt != pixmap->getSize()) {
        Ptr<GfxPixmap> n = new GfxPixmap(pt.x, pt.y, *pixmap);
        blitIndexed(*n, *pixmap, 0, 0);
        pixmap = n;
        drawWidget();
    }
}

/** Blit an image onto a canvas, preserving color indexes.
    When editing indexed pictures, it is pretty important to preserve color
    indexes, even when multiple indexes have the same effective color.
    Therefore, we copy these images almost verbatim.

    Note that this will mess up ForeignPaletteColor images. */
void
blitIndexed(GfxCanvas& dst, GfxPixmap& src, int x, int y)
{
    if (src.getBitsPerPixel() <= 8 && dst.getBitsPerPixel() <= 8) {
        // Pasting 8 bits on 8 bits. Keep color indexes.
        const GfxPoint size = src.getSize();
        const uint32_t ckey = src.getColorKey();
        for (int yy = 0; yy < size.y; ++yy) {
            for (int xx = 0; xx < size.x; ++xx) {
                const uint32_t color = src.getPixel(GfxPoint(xx, yy));
                if (color != ckey || ckey == GFX_TRANSPARENT) {
                    dst.drawPixel(GfxPoint(xx + x, yy + y), color, GFX_ALPHA_OPAQUE);
                }
            }
        }
    } else {
        // Mixed-color or true-color blit
        src.blit(dst, x, y);
    }
}
