/**
  *  \file ui/richdocument.cc
  *  \brief Rich Document Display
  *
  *  (c) 2001-2012 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <climits>
#include "ui/richdocument.h"
#include "ui/uiglobal.h"
#include "gfx/gfxcontext.h"
#include "gfx/clip.h"
#include "gfx/palette.h"
#include "cpluslib/assert.h"

static void
drawKeycap(GfxContext& ctx, int x, int y, int he, int wi)
{
    /* Adjust to leave some inter-line space */
    he -= he/8;

    /* Main frame */
    uint32_t oldColor = ctx.getRawColor();
    if (ui_screen_bits > 8) {
        ctx.setAlpha(128);
    } else {
        ctx.setRawColor(standard_colors[COLOR_DARK]);
    }
    drawHLine(ctx, x+2,    y,      x+wi-3);
    drawHLine(ctx, x+2,    y+he-1, x+wi-3);
    drawVLine(ctx, x,      y+2,    y+he-3);
    drawVLine(ctx, x+wi-1, y+2,    y+he-3);
    drawPixel(ctx, GfxPoint(x+1, y+1));
    drawPixel(ctx, GfxPoint(x+1, y+he-2));
    drawPixel(ctx, GfxPoint(x+wi-2, y+1));
    drawPixel(ctx, GfxPoint(x+wi-2, y+he-2));
    ctx.setRawColor(oldColor);
    ctx.setAlpha(GFX_ALPHA_OPAQUE);

    /* Shadow */
    drawHLine(ctx, x+3,  y+he, x+wi-2);
    drawVLine(ctx, x+wi, y+3,  y+he-2);
    drawPixel(ctx, GfxPoint(x+wi-2, y+he-1));
    drawPixel(ctx, GfxPoint(x+wi-1, y+he-1));
    drawPixel(ctx, GfxPoint(x+wi-1, y+he-2));
}

/**************************** RichTextSplitter ***************************/

namespace {
    /** Splits up RichText into individual per-chunk attributes,
        and generates appropriate RichText::Item's. */
    class RichTextSplitter : public RichTextVisitor {
        ptr_vector<RichDocument::Item>& items;
        int nbold;
        int nbig;
        int nfixed;
        int nunder;
        int nlink;
        int nkey;
        std::vector<UIColor::Color> colors;

        void process(const RichTextAttribute* att, int delta);
     public:
        RichTextSplitter(ptr_vector<RichDocument::Item>& items);
        bool handleText(string_t text);
        bool startAttribute(const RichTextAttribute* att);
        bool endAttribute(const RichTextAttribute* att);
    };
}


RichTextSplitter::RichTextSplitter(ptr_vector<RichDocument::Item>& items)
    : items(items),
      nbold(0), nbig(0), nfixed(0), nunder(0), nlink(0), nkey(0)
{ }

bool
RichTextSplitter::handleText(string_t text)
{
    int add = (nkey ? 7 : 0);
    int font = (nfixed
                ? FONT_FIXED
                : nbig
                ? nbold
                ? FONT_TITLE_BOLD
                : FONT_TITLE
                : nbold
                ? FONT_NORMAL_BOLD
                : FONT_NORMAL);
    UIColor::Color color = !colors.empty() ? colors.back() : UIColor::tc_Static;
    items.push_back(new RichDocument::Item(nlink ? RichDocument::Item::Link : RichDocument::Item::Normal, 0, 0,
                                           fonts[font]->getTextWidth(text) + add,
                                           font, color,
                                           nunder > 0,
                                           nkey > 0,
                                           text));
    return true;
}

bool
RichTextSplitter::startAttribute(const RichTextAttribute* att)
{
    process(att, +1);
    return true;
}

bool
RichTextSplitter::endAttribute(const RichTextAttribute* att)
{
    process(att, -1);
    return true;
}

void
RichTextSplitter::process(const RichTextAttribute* att, int delta)
{
    if (const RichTextStyleAttribute* sa = dynamic_cast<const RichTextStyleAttribute*>(att)) {
        switch (sa->getStyle()) {
         case RichTextStyleAttribute::Bold:
            nbold += delta;
            break;
         case RichTextStyleAttribute::Big:
            nbig += delta;
            break;
         case RichTextStyleAttribute::Fixed:
            nfixed += delta;
            break;
         case RichTextStyleAttribute::Underline:
            nunder += delta;
            break;
         case RichTextStyleAttribute::Key:
            nkey += delta;
            break;
         default:
            break;
        }
    } else if (const RichTextLinkAttribute* la = dynamic_cast<const RichTextLinkAttribute*>(att)) {
        if (delta > 0) {
            items.push_back(new RichDocument::Item(RichDocument::Item::LinkTarget, 0, 0, 0, 0,
                                                   UIColor::tc_Static, false, false, la->getTarget()));
        }
        nlink += delta;
    } else if (const RichTextColorAttribute* ca = dynamic_cast<const RichTextColorAttribute*>(att)) {
        if (delta > 0)
            colors.push_back(ca->getColor());
        else
            colors.pop_back();
    }
}



/****************************** RichDocument *****************************/

/** Marker meaning "no link". */
const RichDocument::linkid_t RichDocument::nil;


/** Default constructor. Makes a blank document. */
RichDocument::RichDocument()
    : content(),
      last_chunk(),
      x(0),
      y(0),
      first_this_line(0),
      page_width(100),
      left_margin(0),
      right_margin(0)
{
    bo_index[0] = bo_index[1] = 0;
    bo_width[0] = bo_width[1] = 0;
    bo_height[0] = bo_height[1] = 0;
}

/** Destructor. */
RichDocument::~RichDocument()
{ }

/** Clear this document. Resets everything except for the page width. */
void
RichDocument::clear()
{
    content.clear();
    last_chunk.clear();
    block_objs.clear();
    x = y = 0;
    first_this_line = 0,
    left_margin = right_margin = 0;
    bo_index[0] = bo_index[1] = 0;
    bo_width[0] = bo_width[1] = 0;
    bo_height[0] = bo_height[1] = 0;
}

/** Set this document's left margin.
    \param lm Left margin in pixels. */
void
RichDocument::setLeftMargin(int lm)
{
    if (x == left_margin + bo_width[BlockObject::Left])
        x = lm + bo_width[BlockObject::Left];
    left_margin = lm;
}

/** Add some text.
    \param text Text to add. */
void
RichDocument::add(const RichText& text)
{
    string_t::size_type start = 0;
    string_t::size_type end;
    while ((end = text.find('\n', start)) != string_t::npos) {
        RichTextSplitter(last_chunk).visit(RichText(text, start, end-start));
        process();
        addNewline();
        start = end+1;
    }

    if (start) {
        RichTextSplitter(last_chunk).visit(RichText(text, start));
    } else {
        RichTextSplitter(last_chunk).visit(text);
    }
    process();
}

/** Add some text.
    \param text Text to add. */
void
RichDocument::add(const string_t& text)
{
    // This can be done more efficient, but this way it's simple:
    add(RichText(text));
}

/** Add some text.
    \param text Text to add. */
void
RichDocument::add(const char* text)
{
    // This can be done more efficient, but this way it's simple:
    add(RichText(text));
}

/** Add newline. Think of this like a HTML "br" tag. */
void
RichDocument::addNewline()
{
    flushWord();
    if (first_this_line == content.size()) {
        /* This line is empty, so just add some space */
        addY(font_heights[FONT_NORMAL]/2);
    } else {
        /* Regular newline */
        flushLine();
    }
}

/** Add new paragraph. Think of this like a HTML "p" tag. */
void
RichDocument::addParagraph()
{
    addNewline();
    addNewline();
}

/** Add text at horizontal position. The position may be outside the
    margin. Further text will be produced after the just-added text.
    This can be used to produce bullets. If the position is before the
    current output position, this starts a new line.
    \param x Position
    \param text Text to add */
void
RichDocument::addAt(int x, const RichText& text)
{
    tabTo(x);
    add(text);
}

/** Add right-justified column text. */
void
RichDocument::addRight(int x, const RichText& text)
{
    /* Write pending previous output */
    flushWord();

    /* Set right margin to our target position */
    int oldRightMargin = right_margin;
    right_margin = page_width - x - bo_width[BlockObject::Right];

    /* Write text, and remember where it ended up */
    size_t start = content.size();
    add(text);
    flushWord();
    size_t end = content.size();

    /* Restore old margin */
    right_margin = oldRightMargin;

    /* If we have produced some output, and that's left of our X, move it right */
    if (end > start && content[end-1]->x + content[end-1]->w < x) {
        int delta = content[end-1]->x + content[end-1]->w - x;
        while (--end >= start) {
            content[end]->x -= delta;
        }
        this->x = x;
    }
}

/** Add centered text. */
void
RichDocument::addCentered(int x, const RichText& text)
{
    /* Write pending previous output */
    flushWord();

    /* Write text, and remember where it ended up */
    size_t start = content.size();
    int startX = this->x;
    add(text);
    flushWord();
    size_t end = content.size();
    int endX = this->x;

    /* If we have produced some output, adjust its position */
    if (end > start && endX > startX && endX < 2*x - startX) {
        int width = endX - startX;
        int delta = x - startX - (width / 2);
        while (end > start) {
            --end;
            content[end]->x += delta;
        }
        this->x += delta;
    }
}

/** Add preformatted text. */
void
RichDocument::addPreformatted(const RichText& text)
{
    /* Write pending previous output */
    tabTo(left_margin);

    /* Set page width to infinity */
    int oldWidth = page_width;
    int oldRightMargin = right_margin;
    page_width = INT_MAX;
    right_margin = 0;

    /* Add output */
    add(text);

    /* Restore margins */
    page_width = oldWidth;
    right_margin = oldRightMargin;
    tabTo(left_margin);
}

/** Add floating object. The object will be placed at the left or
    right border. Text flows around it. */
void
RichDocument::addFloatObject(std::auto_ptr<RichDocumentBlockObject> obj, bool left)
{
    CTASSERT(int(true)  == int(BlockObject::Left));
    CTASSERT(int(false) == int(BlockObject::Right));

    /* Store the object */
    block_objs.push_back(new BlockObject(left ? BlockObject::Left : BlockObject::Right, obj));

    /* When we're at the beginning of the line, try starting the object immediately */
    if (bo_height[left] == 0 && last_chunk.size() == 0 && x == left_margin + bo_width[BlockObject::Left]) {
        bo_index[left] = block_objs.size()-1;
        startNextObject(left);
        x = left_margin + bo_width[BlockObject::Left];
    }
}

/** Add centered object. The object will be placed in the center of
    the page. It will interrupt the text flow. */
void
RichDocument::addCenterObject(std::auto_ptr<RichDocumentBlockObject> obj)
{
    /* Finish current line */
    flushWord();
    flushLine();

    /* Place the object */
    std::auto_ptr<BlockObject> p(new BlockObject(BlockObject::Center, obj));
    p->pos.x = (page_width - p->pos.w) / 2;
    p->pos.y = this->y;
    addY(p->pos.h);
    block_objs.push_back(p.release());
}

/** Move to horizontal position. Further text will be produced
    starting here. The position may be outside the margin. If the
    position is before the current output position, this starts a new
    line.
    \param x Position */
void
RichDocument::tabTo(int x)
{
    x += bo_width[BlockObject::Left];
    flushWord();
    if (this->x > x && first_this_line != content.size()) {
        /* This line already has content, and that is beyond x */
        flushLine();
    }
    this->x = x;
}

/** Finish this document. */
void
RichDocument::finish()
{
    /* Finish text */
    flushWord();
    flushLine();

    /* Finish all floats */
    while (1) {
        if (bo_height[0] > 0) {
            addY(bo_height[0]);
        } else if (bo_height[1] > 0) {
            addY(bo_height[1]);
        } else {
            break;
        }
    }
}

/** Draw document.
    \param ctx Context
    \param area Area to draw in
    \param skipY Skip this many pixels from the top of the document */
void
RichDocument::draw(GfxContext& ctx, const GfxRect area, const int skipY)
{
    for (std::size_t i = 0; i < content.size(); ++i) {
        Item& it = *content[i];
        if (it.kind != Item::LinkTarget) {
            GfxRect pos(it.x, it.y, it.w, font_heights[it.font]);
            pos.x += area.x;
            pos.y += area.y;
            pos.y -= skipY;
            if (pos.isIntersecting(area)) {
                ctx.useFont(*fonts[it.font]);
                if (it.kind & Item::Link) {
                    --pos.h;
                    if (it.kind == Item::LinkFocus) {
                        drawSolidBar(ctx, pos, UIColor::tc_LinkFocus);
                    } else if (it.kind == Item::LinkHover) {
                        drawSolidBar(ctx, pos, UIColor::tc_LinkShade);
                    }
                    ctx.setColor(UIColor::tc_Link);
                    ctx.setLinePattern(0x55 << (pos.x&1));
                    drawHLine(ctx, pos.x, pos.y + font_heights[it.font]*17/20, pos.x+pos.w-1);
                    ctx.setLinePattern(SOLID_LINE);
                    if (it.color != UIColor::tc_Static) {
                        ctx.setColor(it.color);
                    }
                    outText(ctx, pos.x, pos.y, it.text);
                } else {
                    ctx.setColor(it.color);
                    if (it.key) {
                        drawKeycap(ctx, pos.x+1, pos.y, font_heights[it.font], pos.w-3);
                        outText(ctx, pos.x+3, pos.y, it.text);
                    } else {
                        if (it.underline)
                            drawHLine(ctx, pos.x, pos.y + font_heights[it.font]*17/20, pos.x+pos.w-1);
                        outText(ctx, pos.x, pos.y, it.text);
                    }
                }
            }
        }
    }
    for (std::size_t i = 0; i < block_objs.size(); ++i) {
        BlockObject& obj = *block_objs[i];
        GfxRect pos(obj.pos);
        pos.x += area.x;
        pos.y += area.y;
        pos.y -= skipY;
        if (pos.isIntersecting(area)) {
            obj.obj->draw(ctx, pos);
        }
    }
}

/** Get link, given a position. Locates a link that appears at the specified
    position and returns its Id, an opaque value usable to query information
    about the link.
    \param pt Point to look up
    \return Link Id for that point, nil if none */
RichDocument::linkid_t
RichDocument::getLinkFromPos(GfxPoint pt) const
{
    for (std::size_t i = 0; i < content.size(); ++i) {
        if ((content[i]->kind & Item::Link) != 0
            && GfxRect(content[i]->x, content[i]->y, content[i]->w, font_heights[content[i]->font]).contains(pt))
        {
            while (i > 0 && content[i]->kind != Item::LinkTarget)
                --i;
            return i;
        }
    }
    return nil;
}

/** Get link target (parameter to RichTextLinkAttribute) for a link. */
string_t
RichDocument::getLinkTarget(linkid_t link) const
{
    return content[link]->text;
}

/** Change kind (=status) of a link.
    \param link Link Id
    \param kind New link status (must be one of Item::Link, Item::LinkHover, Item::LinkFocus),
                otherwise the link will be destroyed. */
void
RichDocument::setLinkKind(linkid_t link, Item::Kind kind)
{
    while (++link < content.size() && (content[link]->kind & Item::Link) != 0)
        content[link]->kind = kind;
}

/** Get next link.
    \param id Link identifier. nil to get the first link.
    \return Link id of found link, nil if none */
RichDocument::linkid_t
RichDocument::getNextLink(linkid_t id) const
{
    return getNextLink(id, GfxRect(0, 0, 0xFFF0, 0xFFF0));
}

/** Get next link within range.
    \param id Link identifier. nil to get the first link.
    \param limit Return only links within this range
    \return Link id of found link, nil if none */
RichDocument::linkid_t
RichDocument::getNextLink(linkid_t id, GfxRect limit) const
{
    linkid_t pos = (id == nil ? 0 : id+1);
    while (pos < content.size()) {
        if (content[pos]->kind == Item::LinkTarget && isLinkVisible(pos, limit)) {
            return pos;
        }
        ++pos;
    }
    return nil;
}

/** Get previous link.
    \param id Link identifier. nil to get the last link.
    \return Link id of found link, nil if none */
RichDocument::linkid_t
RichDocument::getPreviousLink(linkid_t id) const
{
    return getPreviousLink(id, GfxRect(0, 0, 0xFFF0, 0xFFF0));
}

/** Get previous link within range.
    \param id Link identifier. nil to get the last link.
    \param limit Return only links within this range
    \return Link id of found link, nil if none */
RichDocument::linkid_t
RichDocument::getPreviousLink(linkid_t id, GfxRect limit) const
{
    linkid_t pos = (id == nil ? content.size() : id);
    while (pos > 0) {
        --pos;
        if (content[pos]->kind == Item::LinkTarget && isLinkVisible(pos, limit)) {
            return pos;
        }
    }
    return nil;
}

/** Check whether a link is visible.
    \param id Link identifier, non-nil
    \param limit Range to check
    \return true iff at least a part of this link is visible */
bool
RichDocument::isLinkVisible(linkid_t id, GfxRect limit) const
{
    while (++id < content.size() && (content[id]->kind & Item::Link) != 0)
        if (GfxRect(content[id]->x, content[id]->y,
                    content[id]->w, font_heights[content[id]->font]).isIntersecting(limit))
            return true;
    return false;
}

/** Process pending input. */
void
RichDocument::process()
{
    /* Word wrap. Can this be written more elegantly? */
    std::size_t i = 0;
    std::size_t lastWithSpace = 0;
    int addX = x;
    while (i < last_chunk.size()) {
        int maxX = page_width - right_margin - bo_width[BlockObject::Right];
        addX += last_chunk[i++]->w;
        if (addX > maxX) {
            /* Overflow! Try to find a break point in any chunk here. */
            std::size_t breakItem, breakChar;
            int remX = addX;
            for (breakItem = i; breakItem > 0; --breakItem) {
                Item& it = *last_chunk[breakItem-1];
                remX -= it.w;
                if (it.w != 0) {
                    for (breakChar = it.text.size(); breakChar > 0; --breakChar) {
                        if (it.text[breakChar-1] == ' ' && remX + fonts[it.font]->getTextWidth(it.text.substr(0, breakChar-1)) <= maxX) {
                            /* Found a break point */
                            goto out;
                        }
                    }
                }
            }
         out:
            if (breakItem != 0) {
                /* Means: we found a possible break point.
                   Items [0, breakItem-2] can be copied as is.
                   Item breakItem-1 must be split at breakChar-2. */
                flushItems(breakItem-1);
                last_chunk.erase(last_chunk.begin(), last_chunk.begin() + breakItem-1);

                /* last_chunk[0] now is what originally was breakItem-1 */
                Item& it = *last_chunk[0];
                content.push_back(new Item(it.kind, x, 0, 42, it.font, it.color, it.underline, it.key, it.text.substr(0, breakChar-1)));
                content.back()->w = fonts[it.font]->getTextWidth(content.back()->text);
                flushLine();

                it.text.erase(0, breakChar);
                it.w = fonts[it.font]->getTextWidth(it.text);
                i = 0;
                addX = x;
                lastWithSpace = 0;
            } else {
                /* No break point found. If we are at the beginning of a line,
                   give up: copy the line even though it does not fit.
                   Otherwise, move to a new line and retry. */
                if (x == left_margin + bo_width[BlockObject::Left]) {
                    flushItems(i);
                    last_chunk.erase(last_chunk.begin(), last_chunk.begin() + i);
                }
                flushLine();
                i = 0;
                addX = x;
                lastWithSpace = 0;
            }
        } else {
            /* No overflow, keep going */
            if (last_chunk[i-1]->text.find(' ') != string_t::npos) {
                lastWithSpace = i-1;
            }
        }
    }

    if (lastWithSpace) {
        flushItems(i);
        last_chunk.erase(last_chunk.begin(), last_chunk.begin() + i);
    }
}

/** Add vertical space. This will consume float objects. */
void
RichDocument::addY(int dy)
{
    while (dy > 0) {
        /* Advance to the end of the next float object, if possible. */
        int now = dy;
        for (int side = 0; side < 2; ++side) {
            if (bo_height[side] > 0 && now > bo_height[side]) {
                now = bo_height[side];
            }
        }
        this->y += now;
        dy -= now;

        /* Account for float objects */
        for (int side = 0; side < 2; ++side) {
            if (bo_height[side] > 0) {
                /* Account delta */
                bo_height[side] -= now;

                /* When at end of the object, AND there is another object
                   which is the same width or narrower than the existing one,
                   start it immediately. */
                if (bo_height[side] == 0) {
                    findNextObject(side);
                    if (bo_index[side] < block_objs.size() && block_objs[bo_index[side]]->pos.w <= bo_width[side]) {
                        startNextObject(side);
                    } else {
                        bo_width[side] = 0;
                    }
                }
            }
        }
    }

    /* We can now start new objects at will */
    for (int side = 0; side < 2; ++side) {
        if (bo_height[side] == 0) {
            findNextObject(side);
            if (bo_index[side] < block_objs.size()) {
                startNextObject(side);
            }
        }
    }
}

/** Advance index of specified side to next possible object for that side.
    If there is no such object, the pointer is placed at the end of the
    block_objs list. */
void
RichDocument::findNextObject(int side)
{
    while (bo_index[side] < block_objs.size() && block_objs[bo_index[side]]->kind != side) {
        ++bo_index[side];
    }
}

/** Place the current object of the given side.
    \pre bo_index[side] points to an object */
void
RichDocument::startNextObject(int side)
{
    BlockObject& bo = *block_objs[bo_index[side]];
    bo.pos.x = (side == BlockObject::Left ? 0 : page_width - bo.pos.w);
    bo.pos.y = y;
    bo_height[side] = bo.pos.h;
    bo_width[side] = bo.pos.w;
    ++bo_index[side];
}

/** Finish a line. Makes sure that content contains only fully-rendered text. */
void
RichDocument::flushLine()
{
    /* Figure out maximum height of last line */
    int maxHeight = 0;
    for (std::size_t i = first_this_line; i < content.size(); ++i) {
        Item& it = *content[i];
        if (it.kind != Item::LinkTarget) {
            int thisHeight = font_heights[it.font];
            if (thisHeight > maxHeight)
                maxHeight = thisHeight;
        }
    }

    /* Adjust everything. */
    /* Brute-force approach of baseline handling. Our default fonts have:
          height      10  16  22
          baseline     2   4   5  (=pixels for descenders; "room below a")
       This yields an approximation of height/4 for the baseline height.
       So let's assume a default of 4, and shift the others around a little. */
    for (std::size_t i = first_this_line; i < content.size(); ++i) {
        Item& it = *content[i];
        it.y = y + maxHeight - font_heights[it.font] - 4 + font_heights[it.font]/4;
    }

    /* Advance cursors */
    first_this_line = content.size();
    addY(maxHeight);
    x = left_margin + bo_width[BlockObject::Left];
}

void
RichDocument::flushWord()
{
    flushItems(last_chunk.size());
    last_chunk.clear();
}

void
RichDocument::flushItems(std::size_t n)
{
    for (std::size_t i = 0; i < n; ++i) {
        Item* it = last_chunk.extract_elem(i);
        it->x = x;
        content.push_back(it);
        x += it->w;
    }
}

/***************************** UIRichDocument ****************************/

UIRichDocument::UIRichDocument(GfxPoint pref_size, uint16_t key_flags)
    : UIScrollable(0),
      pref_size(pref_size),
      key_flags(key_flags),
      mdown(false),
      selected_link(RichDocument::nil),
      hover_link(RichDocument::nil)
{ }

void
UIRichDocument::drawContent(GfxCanvas& can)
{
    GfxClipRectFilter filter(getExtent(), can);
    GfxContext ctx;
    ctx.useCanvas(filter);
    ctx.useColorScheme(getSkin());
    drawBackground(ctx, getExtent());
    doc.draw(ctx, getExtent(), getTopY());
    can.update(getExtent());
}

void
UIRichDocument::getLayoutInfo(LayoutInfo& info)
{
    info.min_size = info.pref_size = pref_size;
    info.flags = li_Flexible;
}

void
UIRichDocument::onResize(GfxRect oldSize, bool byLayout)
{
    doc.setPageWidth(getExtent().w);
    updateScrollable();
    UIWidget::onResize(oldSize, byLayout);
}

bool
UIRichDocument::handleEvent(const UIEvent& event, bool second_pass)
{
    if (event.type == ev_Mouse) {
        if (isInside(GfxPoint(event.mouse.x, event.mouse.y))) {
            setSelectedWidget();
            RichDocument::linkid_t link =
                doc.getLinkFromPos(GfxPoint(event.mouse.x - getExtent().x,
                                            event.mouse.y - getExtent().y + getTopY()));
            if (event.mouse.b != 0) {
                if (link != RichDocument::nil)
                    setSelectedLink(link);
                setHoverLink(RichDocument::nil);
                mdown = true;
            } else {
                if (mdown && selected_link != RichDocument::nil && selected_link == link) {
                    sig_link_click.raise(doc.getLinkTarget(selected_link));
                }
                setHoverLink(link);
                mdown = false;
            }
            return true;
        }

        if (mdown && !event.mouse.b) {
            /* Mouse released outside the widget: just reset state */
            mdown = false;
        }
    } else if (event.type == ev_Key) {
        if ((event.key.code == ' ' && (key_flags & fl_Space) != 0)
            || (event.key.code == SDLK_RETURN && (key_flags & fl_Return) != 0))
        {
            /* Return pressed */
            setSelectedWidget();
            mdown = false;
            if (selected_link != RichDocument::nil) {
                sig_link_click.raise(doc.getLinkTarget(selected_link));
            }
            return true;
        }
        if (event.key.code == SDLK_TAB && (key_flags & fl_Tab) != 0) {
            /* Next link */
            setSelectedWidget();
            mdown = false;
            GfxRect area(0, getTopY(), getExtent().w, getExtent().h);
            RichDocument::linkid_t link = doc.getNextLink(selected_link, area);
            if (link == RichDocument::nil && selected_link != RichDocument::nil)
                link = doc.getNextLink(RichDocument::nil, area);
            setSelectedLink(link);
            return true;
        }
        if (event.key.code == ss_Shift + SDLK_TAB && (key_flags & fl_Tab) != 0) {
            /* Previous link */
            setSelectedWidget();
            mdown = false;
            GfxRect area(0, getTopY(), getExtent().w, getExtent().h);
            RichDocument::linkid_t link = doc.getPreviousLink(selected_link, area);
            if (link == RichDocument::nil && selected_link != RichDocument::nil)
                link = doc.getPreviousLink(RichDocument::nil, area);
            setSelectedLink(link);
            return true;
        }

        if ((key_flags & fl_Scroll) != 0) {
            switch (event.key.code) {
             case SDLK_UP:
             case vk_WheelUp:
                scrollTo(getCurrent()-1, -1);
                return true;
             case SDLK_DOWN:
             case vk_WheelDown:
                scrollTo(getCurrent()+1, 1);
                return true;
             case SDLK_PAGEUP:
                scrollTo(getCurrent()-getExtent().h/getScrollStep(), 1);
                return true;
             case SDLK_PAGEDOWN:
             case ' ':
                scrollTo(getCurrent()+getExtent().h/getScrollStep(), 1);
                return true;
             case SDLK_PAGEUP + ss_Ctrl:
             case SDLK_HOME + ss_Ctrl:
             case '<':
                scrollTo(0, 1);
                return true;
             case SDLK_PAGEDOWN + ss_Ctrl:
             case SDLK_END + ss_Ctrl:
             case '>':
                scrollTo(getSize(), 1);
                return true;
            }
        }
    }
    return UIWidget::handleEvent(event, second_pass);
}

void
UIRichDocument::onStateChange(int state, bool enabled)
{
    if ((state & st_Selected) != 0 && !enabled) {
        setHoverLink(RichDocument::nil);
        mdown = false;
    }
    UIWidget::onStateChange(state, enabled);
}

/** Handle document update. Must be called when the document was updated
    while the widget is displayed. */
void
UIRichDocument::handleDocumentUpdate()
{
    // Configure UIScrollable
    updateScrollable();

    // Configure rest
    selected_link = hover_link = RichDocument::nil;
    drawWidget();
    postMouseEvent();
}

void
UIRichDocument::scrollTo(int pos, int /*corr*/)
{
    if (pos >= getSize()) {
        pos = getSize() - 1;
    }
    if (pos < 0) {
        pos = 0;
    }
    if (pos != getCurrent()) {
        setCurrent(pos);
        setTop(pos);
        if (selected_link != RichDocument::nil && !doc.isLinkVisible(selected_link, GfxRect(0, getTopY(), getExtent().w, getExtent().h))) {
            setSelectedLink(RichDocument::nil);
        }
        if (hover_link != RichDocument::nil && !doc.isLinkVisible(hover_link, GfxRect(0, getTopY(), getExtent().w, getExtent().h))) {
            setHoverLink(RichDocument::nil);
        }
        drawWidget();
        sig_moved.raise(getId());
    }
}

bool
UIRichDocument::isAccessible(int /*pos*/)
{
    return true;
}

void
UIRichDocument::setSelectedLink(RichDocument::linkid_t link)
{
    if (link != selected_link) {
        RichDocument::linkid_t old = selected_link;
        selected_link = link;
        setLink(old);
        setLink(selected_link);
        drawWidget();
    }
}

void
UIRichDocument::setHoverLink(RichDocument::linkid_t link)
{
    if (link != hover_link) {
        RichDocument::linkid_t old = hover_link;
        hover_link = link;
        setLink(old);
        setLink(hover_link);
        drawWidget();
    }
}

void
UIRichDocument::setLink(RichDocument::linkid_t link)
{
    if (link != RichDocument::nil) {
        doc.setLinkKind(link, (link == selected_link ? RichDocument::Item::LinkFocus
                               : link == hover_link ? RichDocument::Item::LinkHover
                               : RichDocument::Item::Link));
    }
}

int
UIRichDocument::getTopY() const
{
    return getCurrent() * getScrollStep();
}

int
UIRichDocument::getScrollStep() const
{
    int n = font_heights[FONT_NORMAL];
    if (n <= 0) {
        n = 1;
    }
    return n;
}

void
UIRichDocument::updateScrollable()
{
    int ss = getScrollStep();
    int dh = doc.getDocumentHeight() / ss + 2;
    int ps = (getExtent().h + ss-1) / ss;
    if (dh < 0) {
        dh = ss;
    }

    setCurrent(0);
    setTop(0);
    setPageSize(1);
    setSize(dh - ps);
    sig_moved.raise(getId());
}
